#include "StdAfx.h"

using namespace System;
using namespace System::Collections;
using namespace System::Collections::Generic;
using namespace OpenTK;
using namespace OpenTK::Graphics;
using namespace OpenTK::Graphics::OpenGL;

#include "NitroROM.h"
#include "NitroFile.h"
#include "NitroOverlay.h"
#include "BMD.h"
#include "levelclasses.h"
#include "3dstuff.h"


void ClampRotation(float% val, float twopi)
{
	if (val > twopi)
	{
		while (val > twopi)
			val -= twopi;
	}
	else if (val < -twopi)
	{
		while (val < -twopi)
			val += twopi;
	}
}


CGLG_BMDPart::CGLG_BMDPart(float scale, CBMD::CMatPolyPair^ mppair)
{
	m_NumVertices = 0;
	m_Translucent = mppair->IsTranslucent();

	UInt32 alpha = ((mppair->m_PolyAttributes << 11) & 0xF8000000) | ((mppair->m_PolyAttributes << 6) & 0x07000000);

	switch (mppair->m_PolyAttributes & 0x30)
	{
	case 0x00: m_TexEnvMode = TextureEnvMode::Modulate; break;
	case 0x10: m_TexEnvMode = TextureEnvMode::Decal; break;
//	case 0x20: throw gcnew Exception("TOON/HIGHLIGHT SHADING"); break;
//	case 0x30: throw gcnew Exception("SHADOW POLYGON"); break;
	default: m_TexEnvMode = TextureEnvMode::Modulate; break;
	}

	List<float>^ vertices = gcnew List<float>();
	List<UInt32>^ colors = gcnew List<UInt32>();
	List<UInt32>^ pcolors = gcnew List<UInt32>();
	List<float>^ texcoords = gcnew List<float>();
	List<UInt16>^ indices = gcnew List<UInt16>();
	UInt16 curindex = 0;

#define ADDVTX(svtx) \
vertices->Add(svtx.m_Position.X * scale); vertices->Add(svtx.m_Position.Y * scale); vertices->Add(svtx.m_Position.Z * scale); \
colors->Add(svtx.m_Color | alpha); \
pcolors->Add(0xF0000000); \
texcoords->Add(svtx.m_Texcoord.X); texcoords->Add(svtx.m_Texcoord.Y);

	for each (CBMD::CGeometryBlock^ gblock in mppair->m_Geometry)
	{
		// translate geometry block into triangles
		switch (gblock->m_Type)
		{
		case 0: // triangles
			{
				for each (CBMD::SVertex vtx in gblock->m_VertexList)
				{
					ADDVTX(vtx);
					indices->Add(curindex);
					curindex++;
				}
			}
			break;

		case 1: // quads
			{
				for (int j = 0; j < gblock->m_VertexList->Count; j += 4)
				{
					CBMD::SVertex vtx1 = gblock->m_VertexList[j  ];
					CBMD::SVertex vtx2 = gblock->m_VertexList[j+1];
					CBMD::SVertex vtx3 = gblock->m_VertexList[j+2];
					CBMD::SVertex vtx4 = gblock->m_VertexList[j+3];
					ADDVTX(vtx1);
					ADDVTX(vtx2);
					ADDVTX(vtx3);
					ADDVTX(vtx4);

					indices->Add(curindex  );
					indices->Add(curindex+1);
					indices->Add(curindex+2);
					indices->Add(curindex  );
					indices->Add(curindex+2);
					indices->Add(curindex+3);
					curindex += 4;
				}
			}
			break;

		case 2: // triangle strip
			{
				CBMD::SVertex svtx1 = gblock->m_VertexList[0];
				CBMD::SVertex svtx2 = gblock->m_VertexList[1];
				CBMD::SVertex svtx3 = gblock->m_VertexList[2];
				ADDVTX(svtx1);
				ADDVTX(svtx2);
				ADDVTX(svtx3);

				indices->Add(curindex  );
				indices->Add(curindex+1);
				indices->Add(curindex+2);
				curindex += 3;

				for (int j = 3; j < gblock->m_VertexList->Count; j++)
				{
					CBMD::SVertex vtx = gblock->m_VertexList[j];
					ADDVTX(vtx);

					if (j & 0x1)
					{
						indices->Add(curindex-1);
						indices->Add(curindex-2);
						indices->Add(curindex  );
						curindex++;
					}
					else
					{
						indices->Add(curindex-2);
						indices->Add(curindex-1);
						indices->Add(curindex  );
						curindex++;
					}
				}
			}
			break;

		case 3: // quad strip
			{
				CBMD::SVertex svtx1 = gblock->m_VertexList[0];
				CBMD::SVertex svtx2 = gblock->m_VertexList[1];
				CBMD::SVertex svtx3 = gblock->m_VertexList[2];
				CBMD::SVertex svtx4 = gblock->m_VertexList[3];
				ADDVTX(svtx1);
				ADDVTX(svtx2);
				ADDVTX(svtx3);
				ADDVTX(svtx4);

				indices->Add(curindex  );
				indices->Add(curindex+1);
				indices->Add(curindex+2);
				indices->Add(curindex+2);
				indices->Add(curindex+1);
				indices->Add(curindex+3);
				curindex += 4;

				for (int j = 4; j < gblock->m_VertexList->Count; j += 2)
				{
					CBMD::SVertex vtx1 = gblock->m_VertexList[j  ];
					CBMD::SVertex vtx2 = gblock->m_VertexList[j+1];
					ADDVTX(vtx1);
					ADDVTX(vtx2);

					indices->Add(curindex-2);
					indices->Add(curindex-1);
					indices->Add(curindex  );
					indices->Add(curindex  );
					indices->Add(curindex-1);
					indices->Add(curindex+1);
					curindex += 2;
				}
			}
			break;
		}
	}

#undef ADDVTX

	m_NumVertices = indices->Count;

	// send the vertices and indices to video memory
	m_VBOID = gcnew array<UInt32>(5);
	GL::GenBuffers(5, m_VBOID);

	GL::BindBuffer(BufferTarget::ElementArrayBuffer, m_VBOID[0]);
	GL::BufferData(BufferTarget::ElementArrayBuffer, IntPtr((int)(indices->Count * sizeof(UInt16))), indices->ToArray(), BufferUsageHint::StaticDraw);

	GL::BindBuffer(BufferTarget::ArrayBuffer, m_VBOID[1]);
	GL::BufferData(BufferTarget::ArrayBuffer, IntPtr((int)(vertices->Count * sizeof(float))), vertices->ToArray(), BufferUsageHint::StaticDraw);

	GL::BindBuffer(BufferTarget::ArrayBuffer, m_VBOID[2]);
	GL::BufferData(BufferTarget::ArrayBuffer, IntPtr((int)(colors->Count * sizeof(UInt32))), colors->ToArray(), BufferUsageHint::StaticDraw);

	GL::BindBuffer(BufferTarget::ArrayBuffer, m_VBOID[3]);
	GL::BufferData(BufferTarget::ArrayBuffer, IntPtr((int)(pcolors->Count * sizeof(UInt32))), pcolors->ToArray(), BufferUsageHint::StaticDraw);

	GL::BindBuffer(BufferTarget::ArrayBuffer, m_VBOID[4]);
	GL::BufferData(BufferTarget::ArrayBuffer, IntPtr((int)(texcoords->Count * sizeof(float))), texcoords->ToArray(), BufferUsageHint::StaticDraw);

	// texturing!
	// TODO- intelligent texture caching
	if (mppair->m_Texture != nullptr)
	{
		m_TextureID = GL::GenTexture();
		GL::BindTexture(TextureTarget::Texture2D, m_TextureID);
		GL::TexImage2D(TextureTarget::Texture2D, 0, PixelInternalFormat::Four, mppair->m_TexWidth, mppair->m_TexHeight, 0, PixelFormat::Bgra, PixelType::UnsignedByte, mppair->m_Texture);

		GL::TexParameter(TextureTarget::Texture2D, TextureParameterName::TextureMinFilter, (int)TextureMinFilter::Nearest);
		GL::TexParameter(TextureTarget::Texture2D, TextureParameterName::TextureMagFilter, (int)TextureMagFilter::Nearest);

		GL::TexParameter(TextureTarget::Texture2D, TextureParameterName::TextureWrapS, (int)((mppair->m_TexParams & 0x10000) ? ((mppair->m_TexParams & 0x40000) ? TextureWrapMode::MirroredRepeat : TextureWrapMode::Repeat) : TextureWrapMode::Clamp));
		GL::TexParameter(TextureTarget::Texture2D, TextureParameterName::TextureWrapT, (int)((mppair->m_TexParams & 0x20000) ? ((mppair->m_TexParams & 0x80000) ? TextureWrapMode::MirroredRepeat : TextureWrapMode::Repeat) : TextureWrapMode::Clamp));
	}
	else
		m_TextureID = 0;

	delete indices;
	delete texcoords;
	delete pcolors;
	delete colors;
	delete vertices;
}

CGLG_BMDPart::~CGLG_BMDPart()
{
	GL::DeleteBuffers(5, m_VBOID);
	delete m_VBOID;
}

void CGLG_BMDPart::Render(int colorvbo)
{
	if (!m_Enabled) return;

	if (colorvbo == 2)
	{
		GL::BindTexture(TextureTarget::Texture2D, m_TextureID);
		GL::TexEnv(TextureEnvTarget::TextureEnv, TextureEnvParameter::TextureEnvMode, (int)m_TexEnvMode);
	}
	
	GL::BindBuffer(BufferTarget::ArrayBuffer, m_VBOID[1]);
	GL::VertexPointer(3, VertexPointerType::Float, 0, IntPtr::Zero);

	GL::BindBuffer(BufferTarget::ArrayBuffer, m_VBOID[colorvbo]);
	GL::ColorPointer(4, ColorPointerType::UnsignedByte, 0, IntPtr::Zero);

	GL::BindBuffer(BufferTarget::ArrayBuffer, m_VBOID[4]);
	GL::TexCoordPointer(2, TexCoordPointerType::Float, 0, IntPtr::Zero);

	GL::BindBuffer(BufferTarget::ElementArrayBuffer, m_VBOID[0]);
	GL::DrawElements(BeginMode::Triangles, m_NumVertices, DrawElementsType::UnsignedShort, IntPtr::Zero);
}


CGLG_ObjList::CGLG_ObjList(int type, int layer, List<IPositionedObject^>^ objlist)
{
	array<UInt32>^ ccolors = {0xFF400000, 0xFF404000, 0xFF004000};
	array<UInt32>^ bcolors = {0xFFFF0000, 0xFFFFFF00, 0xFF00FF00};

	m_ObjType = type;
	m_ObjLayer = layer;
	m_ObjList = objlist;

	m_CubeColor = ccolors[type];
	m_BorderColor = bcolors[type];

	b_IsEmpty = (objlist->Count == 0);
	m_VBOID = nullptr;

	UpdateAll();
}

CGLG_ObjList::~CGLG_ObjList()
{
	if (m_VBOID)
	{
		GL::DeleteBuffers(4, m_VBOID);
		delete m_VBOID;
	}

	delete m_ObjLineVtxOffsets;
	delete m_ObjTriVtxOffsets;
}

void CGLG_ObjList::Update(int objidx)
{
	UInt32 linevtxoffset = m_ObjLineVtxOffsets[objidx];
	UInt32 trivtxoffset = m_ObjTriVtxOffsets[objidx];
	IPositionedObject^ obj = m_ObjList[objidx];

	List<float>^ vertices = gcnew List<float>();

#define ADDVTX(_x,_y,_z) vertices->Add(_x); vertices->Add(_y); vertices->Add(_z);

	float xpos = obj->XPos;
	float ypos = obj->YPos;
	float zpos = obj->ZPos;

	array<Vector3>^ cube = {
		Vector3(-0.05f, -0.05f, -0.05f),
		Vector3( 0.05f, -0.05f, -0.05f),
		Vector3(-0.05f,  0.05f, -0.05f),
		Vector3(-0.05f, -0.05f,  0.05f),
		Vector3( 0.05f,  0.05f, -0.05f),
		Vector3(-0.05f,  0.05f,  0.05f),
		Vector3( 0.05f, -0.05f,  0.05f),
		Vector3( 0.05f,  0.05f,  0.05f),
	};

	Matrix4 translation = Matrix4::CreateTranslation(xpos, ypos, zpos);

	if (m_ObjList[0]->CanBeRotated())
	{
		float yrot = (float)obj->YRot * Math::PI / 8.f;
		Matrix4 yrotmtx = Matrix4::CreateRotationY(yrot);

		for each (Vector3% vec in cube)
		{
			vec = Vector3::TransformPosition(vec, yrotmtx);
		}
	}

	for each (Vector3% vec in cube)
	{
		vec = Vector3::TransformPosition(vec, translation);
		ADDVTX(vec.X, vec.Y, vec.Z);
	}

	GL::BindBuffer(BufferTarget::ArrayBuffer, m_VBOID[1]);
	GL::BufferSubData(BufferTarget::ArrayBuffer, IntPtr((int)(trivtxoffset * sizeof(float))), IntPtr((int)(vertices->Count * sizeof(float))), vertices->ToArray());

	vertices->Clear();

	if (m_ObjList[0]->CanBeRotated())
	{
		float yrot = (float)obj->YRot * Math::PI / 8.f;
		Matrix4 yrotmtx = Matrix4::CreateRotationY(yrot);

		array<Vector3>^ axes = {
			Vector3(0.f, 0.f, 0.f),
			Vector3(0.1f, 0.f, 0.f),
			Vector3(0.f, 0.f, 0.f),
			Vector3(0.f, 0.1f, 0.f),
			Vector3(0.f, 0.f, 0.f),
			Vector3(0.f, 0.f, 0.1f)
		};

		for each (Vector3% vec in axes)
		{
			vec = Vector3::TransformPosition(vec, yrotmtx);
			vec = Vector3::TransformPosition(vec, translation);
			ADDVTX(vec.X, vec.Y, vec.Z);
		}
	}

	for each (Vector3% vec in cube)
	{
		ADDVTX(vec.X, vec.Y, vec.Z);
	}

	GL::BufferSubData(BufferTarget::ArrayBuffer, IntPtr((int)(linevtxoffset * sizeof(float))), IntPtr((int)(vertices->Count * sizeof(float))), vertices->ToArray());

#undef ADDVTX

	delete vertices;
}

void CGLG_ObjList::UpdateAll()
{
	if (m_ObjTriVtxOffsets) { delete m_ObjTriVtxOffsets; m_ObjTriVtxOffsets = nullptr; }
	if (m_ObjLineVtxOffsets) { delete m_ObjLineVtxOffsets; m_ObjLineVtxOffsets = nullptr; }

	if (m_VBOID)
	{
		GL::DeleteBuffers(4, m_VBOID);
		delete m_VBOID; m_VBOID = nullptr;
	}

	if (b_IsEmpty) return;

	m_ObjTriVtxOffsets = gcnew List<UInt32>(m_ObjList->Count);
	m_ObjLineVtxOffsets = gcnew List<UInt32>(m_ObjList->Count);

	List<float>^ vertices = gcnew List<float>();
	List<UInt32>^ colors = gcnew List<UInt32>();
	List<UInt32>^ pcolors = gcnew List<UInt32>();
	List<UInt16>^ indices = gcnew List<UInt16>();
	UInt16 curindex = 0;

#define ADDVTX(_x,_y,_z,_color) \
vertices->Add(_x); vertices->Add(_y); vertices->Add(_z); \
colors->Add(_color); \
pcolors->Add(((obj->m_UniqueID & 0xFF00FF00)) | ((obj->m_UniqueID & 0x00FF0000) >> 16) | ((obj->m_UniqueID & 0x000000FF) << 16));

	for each (IPositionedObject^ obj in m_ObjList)
	{
		m_ObjTriVtxOffsets->Add(vertices->Count);

		float xpos = obj->XPos;
		float ypos = obj->YPos;
		float zpos = obj->ZPos;

		array<Vector3>^ cube = {
			Vector3(-0.05f, -0.05f, -0.05f),
			Vector3( 0.05f, -0.05f, -0.05f),
			Vector3(-0.05f,  0.05f, -0.05f),
			Vector3(-0.05f, -0.05f,  0.05f),
			Vector3( 0.05f,  0.05f, -0.05f),
			Vector3(-0.05f,  0.05f,  0.05f),
			Vector3( 0.05f, -0.05f,  0.05f),
			Vector3( 0.05f,  0.05f,  0.05f),
		};

		Matrix4 translation = Matrix4::CreateTranslation(xpos, ypos, zpos);

		if (m_ObjList[0]->CanBeRotated())
		{
			float yrot = (float)obj->YRot * Math::PI / 8.f;
			Matrix4 yrotmtx = Matrix4::CreateRotationY(yrot);

			for each (Vector3% vec in cube)
			{
				vec = Vector3::TransformPosition(vec, yrotmtx);
			}
		}

		for each (Vector3% vec in cube)
		{
			vec = Vector3::TransformPosition(vec, translation);
			ADDVTX(vec.X, vec.Y, vec.Z, m_CubeColor);
		}

		indices->Add(curindex+ 4); indices->Add(curindex+ 1); indices->Add(curindex+ 6);
		indices->Add(curindex+ 4); indices->Add(curindex+ 6); indices->Add(curindex+ 7);
		indices->Add(curindex+ 7); indices->Add(curindex+ 6); indices->Add(curindex+ 3);
		indices->Add(curindex+ 7); indices->Add(curindex+ 3); indices->Add(curindex+ 5);
		indices->Add(curindex+ 5); indices->Add(curindex+ 3); indices->Add(curindex   );
		indices->Add(curindex+ 5); indices->Add(curindex   ); indices->Add(curindex+ 2);
		indices->Add(curindex+ 2); indices->Add(curindex   ); indices->Add(curindex+ 1);
		indices->Add(curindex+ 2); indices->Add(curindex+ 1); indices->Add(curindex+ 4);
		indices->Add(curindex+ 4); indices->Add(curindex+ 7); indices->Add(curindex+ 5);
		indices->Add(curindex+ 4); indices->Add(curindex+ 5); indices->Add(curindex+ 2);
		indices->Add(curindex   ); indices->Add(curindex+ 3); indices->Add(curindex+ 6);
		indices->Add(curindex   ); indices->Add(curindex+ 6); indices->Add(curindex+ 1);
		curindex += 8;
	}

	m_NumTriVertices = indices->Count;

	for each (IPositionedObject^ obj in m_ObjList)
	{
		m_ObjLineVtxOffsets->Add(vertices->Count);

		float xpos = obj->XPos;
		float ypos = obj->YPos;
		float zpos = obj->ZPos;

		array<Vector3>^ cube = {
			Vector3(-0.05f, -0.05f, -0.05f),
			Vector3( 0.05f, -0.05f, -0.05f),
			Vector3(-0.05f,  0.05f, -0.05f),
			Vector3(-0.05f, -0.05f,  0.05f),
			Vector3( 0.05f,  0.05f, -0.05f),
			Vector3(-0.05f,  0.05f,  0.05f),
			Vector3( 0.05f, -0.05f,  0.05f),
			Vector3( 0.05f,  0.05f,  0.05f),
		};

		Matrix4 translation = Matrix4::CreateTranslation(xpos, ypos, zpos);

		if (m_ObjList[0]->CanBeRotated())
		{
			float yrot = (float)obj->YRot * Math::PI / 8.f;
			Matrix4 yrotmtx = Matrix4::CreateRotationY(yrot);

			for each (Vector3% vec in cube)
			{
				vec = Vector3::TransformPosition(vec, yrotmtx);
			}

			array<Vector3>^ axes = {
				Vector3(0.f, 0.f, 0.f),
				Vector3(0.1f, 0.f, 0.f),
				Vector3(0.f, 0.f, 0.f),
				Vector3(0.f, 0.1f, 0.f),
				Vector3(0.f, 0.f, 0.f),
				Vector3(0.f, 0.f, 0.1f)
			};
			array<UInt32>^ axescolors = {0xFF0000FF, 0xFF00FF00, 0xFFFF0000};
			int ac = 0;

			for each (Vector3% vec in axes)
			{
				vec = Vector3::TransformPosition(vec, yrotmtx);
				vec = Vector3::TransformPosition(vec, translation);
				ADDVTX(vec.X, vec.Y, vec.Z, axescolors[ac>>1]);
				ac++;
			}

			indices->Add(curindex++);
			indices->Add(curindex++);
			indices->Add(curindex++);
			indices->Add(curindex++);
			indices->Add(curindex++);
			indices->Add(curindex++);
		}

		for each (Vector3% vec in cube)
		{
			vec = Vector3::TransformPosition(vec, translation);
			ADDVTX(vec.X, vec.Y, vec.Z, m_BorderColor);
		}

		indices->Add(curindex+ 4); indices->Add(curindex+ 1); indices->Add(curindex+ 1); indices->Add(curindex+ 6);
		indices->Add(curindex+ 6); indices->Add(curindex+ 7); indices->Add(curindex+ 7); indices->Add(curindex+ 4);
		indices->Add(curindex+ 5); indices->Add(curindex+ 3); indices->Add(curindex+ 3); indices->Add(curindex   );
		indices->Add(curindex   ); indices->Add(curindex+ 2); indices->Add(curindex+ 2); indices->Add(curindex+ 5);
		indices->Add(curindex+ 5); indices->Add(curindex+ 7); indices->Add(curindex+ 2); indices->Add(curindex+ 4);
		indices->Add(curindex+ 3); indices->Add(curindex+ 6); indices->Add(curindex   ); indices->Add(curindex+ 1);
		curindex += 8;
	}

	m_NumLineVertices = indices->Count - m_NumTriVertices;

#undef ADDVTX
	
	// send the vertices and indices to video memory
	m_VBOID = gcnew array<UInt32>(4);
	GL::GenBuffers(4, m_VBOID);

	GL::BindBuffer(BufferTarget::ElementArrayBuffer, m_VBOID[0]);
	GL::BufferData(BufferTarget::ElementArrayBuffer, IntPtr((int)(indices->Count * sizeof(UInt16))), indices->ToArray(), BufferUsageHint::StaticDraw);

	GL::BindBuffer(BufferTarget::ArrayBuffer, m_VBOID[1]);
	GL::BufferData(BufferTarget::ArrayBuffer, IntPtr((int)(vertices->Count * sizeof(float))), vertices->ToArray(), BufferUsageHint::StaticDraw);

	GL::BindBuffer(BufferTarget::ArrayBuffer, m_VBOID[2]);
	GL::BufferData(BufferTarget::ArrayBuffer, IntPtr((int)(colors->Count * sizeof(UInt32))), colors->ToArray(), BufferUsageHint::StaticDraw);

	GL::BindBuffer(BufferTarget::ArrayBuffer, m_VBOID[3]);
	GL::BufferData(BufferTarget::ArrayBuffer, IntPtr((int)(pcolors->Count * sizeof(UInt32))), pcolors->ToArray(), BufferUsageHint::StaticDraw);

	delete indices;
	delete pcolors;
	delete colors;
	delete vertices;
}

void CGLG_ObjList::Render(int colorvbo)
{
	if (!m_Enabled) return;
	if (b_IsEmpty) return;

	GL::DisableClientState(ArrayCap::TextureCoordArray);

	GL::BindTexture(TextureTarget::Texture2D, 0);

	GL::BindBuffer(BufferTarget::ArrayBuffer, m_VBOID[1]);
	GL::VertexPointer(3, VertexPointerType::Float, 0, IntPtr::Zero);

	GL::BindBuffer(BufferTarget::ArrayBuffer, m_VBOID[colorvbo]);
	GL::ColorPointer(4, ColorPointerType::UnsignedByte, 0, IntPtr::Zero);

	GL::BindBuffer(BufferTarget::ElementArrayBuffer, m_VBOID[0]);
	GL::DrawElements(BeginMode::Triangles, m_NumTriVertices, DrawElementsType::UnsignedShort, IntPtr::Zero);
	GL::DrawElements(BeginMode::Lines, m_NumLineVertices, DrawElementsType::UnsignedShort, IntPtr((int)(m_NumTriVertices * sizeof(UInt16))));

	GL::EnableClientState(ArrayCap::TextureCoordArray);
}



CGLG_ObjSelCube::CGLG_ObjSelCube(UInt32 color)
{
	m_SelObject = nullptr;

	List<float>^ vertices = gcnew List<float>();
	List<UInt32>^ colors = gcnew List<UInt32>();
	List<Byte>^ indices = gcnew List<Byte>();

#define ADDVTX(_x,_y,_z,_alpha) \
vertices->Add(_x); vertices->Add(_y); vertices->Add(_z); \
colors->Add(_alpha | color);

	// cube lines
	ADDVTX(-0.0505f, -0.0505f, -0.0505f, 0xFF000000);
	ADDVTX( 0.0505f, -0.0505f, -0.0505f, 0xFF000000);
	ADDVTX(-0.0505f,  0.0505f, -0.0505f, 0xFF000000);
	ADDVTX(-0.0505f, -0.0505f,  0.0505f, 0xFF000000);
	ADDVTX( 0.0505f,  0.0505f, -0.0505f, 0xFF000000);
	ADDVTX(-0.0505f,  0.0505f,  0.0505f, 0xFF000000);
	ADDVTX( 0.0505f, -0.0505f,  0.0505f, 0xFF000000);
	ADDVTX( 0.0505f,  0.0505f,  0.0505f, 0xFF000000);

	indices->Add(4); indices->Add(1); indices->Add(1); indices->Add(6);
	indices->Add(6); indices->Add(7); indices->Add(7); indices->Add(4);
	indices->Add(5); indices->Add(3); indices->Add(3); indices->Add(0);
	indices->Add(0); indices->Add(2); indices->Add(2); indices->Add(5);
	indices->Add(5); indices->Add(7); indices->Add(2); indices->Add(4);
	indices->Add(3); indices->Add(6); indices->Add(0); indices->Add(1);

	// cube faces
	ADDVTX(-0.0505f, -0.0505f, -0.0505f, 0x80000000);
	ADDVTX( 0.0505f, -0.0505f, -0.0505f, 0x80000000);
	ADDVTX(-0.0505f,  0.0505f, -0.0505f, 0x80000000);
	ADDVTX(-0.0505f, -0.0505f,  0.0505f, 0x80000000);
	ADDVTX( 0.0505f,  0.0505f, -0.0505f, 0x80000000);
	ADDVTX(-0.0505f,  0.0505f,  0.0505f, 0x80000000);
	ADDVTX( 0.0505f, -0.0505f,  0.0505f, 0x80000000);
	ADDVTX( 0.0505f,  0.0505f,  0.0505f, 0x80000000);

	indices->Add(8+4); indices->Add(8+1); indices->Add(8+6);
	indices->Add(8+4); indices->Add(8+6); indices->Add(8+7);
	indices->Add(8+7); indices->Add(8+6); indices->Add(8+3);
	indices->Add(8+7); indices->Add(8+3); indices->Add(8+5);
	indices->Add(8+5); indices->Add(8+3); indices->Add(8+0);
	indices->Add(8+5); indices->Add(8+0); indices->Add(8+2);
	indices->Add(8+2); indices->Add(8+0); indices->Add(8+1);
	indices->Add(8+2); indices->Add(8+1); indices->Add(8+4);
	indices->Add(8+4); indices->Add(8+7); indices->Add(8+5);
	indices->Add(8+4); indices->Add(8+5); indices->Add(8+2);
	indices->Add(8+0); indices->Add(8+3); indices->Add(8+6);
	indices->Add(8+0); indices->Add(8+6); indices->Add(8+1);

	// shadow volume
	ADDVTX(-0.505f, -64.f, -0.505f, 0xD0000000);
	ADDVTX( 0.505f, -64.f, -0.505f, 0xD0000000);
	ADDVTX(-0.505f, -64.f,  0.505f, 0xD0000000);
	ADDVTX( 0.505f, -64.f,  0.505f, 0xD0000000);

	indices->Add(11); indices->Add(18); indices->Add(19);
	indices->Add(14); indices->Add(11); indices->Add(19);
	indices->Add(14); indices->Add(19); indices->Add(17);
	indices->Add( 9); indices->Add(14); indices->Add(17);
	indices->Add( 9); indices->Add(17); indices->Add(16);
	indices->Add( 8); indices->Add( 9); indices->Add(16);
	indices->Add( 8); indices->Add(16); indices->Add(18);
	indices->Add(11); indices->Add( 8); indices->Add(18);

#undef ADDVTX
	
	// send the vertices and indices to video memory
	m_VBOID = gcnew array<UInt32>(3);
	GL::GenBuffers(3, m_VBOID);

	GL::BindBuffer(BufferTarget::ElementArrayBuffer, m_VBOID[0]);
	GL::BufferData(BufferTarget::ElementArrayBuffer, IntPtr((int)(indices->Count * sizeof(Byte))), indices->ToArray(), BufferUsageHint::StaticDraw);

	GL::BindBuffer(BufferTarget::ArrayBuffer, m_VBOID[1]);
	GL::BufferData(BufferTarget::ArrayBuffer, IntPtr((int)(vertices->Count * sizeof(float))), vertices->ToArray(), BufferUsageHint::StaticDraw);

	GL::BindBuffer(BufferTarget::ArrayBuffer, m_VBOID[2]);
	GL::BufferData(BufferTarget::ArrayBuffer, IntPtr((int)(colors->Count * sizeof(UInt32))), colors->ToArray(), BufferUsageHint::StaticDraw);

	delete indices;
	delete colors;
	delete vertices;
}

CGLG_ObjSelCube::~CGLG_ObjSelCube()
{
	GL::DeleteBuffers(3, m_VBOID);
	delete m_VBOID;
}

void CGLG_ObjSelCube::Update(IPositionedObject^ sel)
{
	m_SelObject = sel;
	if (m_SelObject == nullptr) return;

	m_PosMatrix = Matrix4::CreateTranslation(m_SelObject->XPos, m_SelObject->YPos, m_SelObject->ZPos);

	if (m_SelObject->GetObjType() == 0)
	{
		m_PosMatrix = Matrix4::Mult(Matrix4::CreateRotationY((float)m_SelObject->YRot * Math::PI / 8.f), m_PosMatrix);
	}
}

void CGLG_ObjSelCube::RenderActually()
{
	if (m_SelObject == nullptr) return;

	GL::DisableClientState(ArrayCap::TextureCoordArray);

	GL::BindTexture(TextureTarget::Texture2D, 0);

	GL::PushMatrix();
	GL::MultMatrix(m_PosMatrix);

	GL::BindBuffer(BufferTarget::ArrayBuffer, m_VBOID[1]);
	GL::VertexPointer(3, VertexPointerType::Float, 0, IntPtr::Zero);

	GL::BindBuffer(BufferTarget::ArrayBuffer, m_VBOID[2]);
	GL::ColorPointer(4, ColorPointerType::UnsignedByte, 0, IntPtr::Zero);

	GL::BindBuffer(BufferTarget::ElementArrayBuffer, m_VBOID[0]);
	GL::DrawElements(BeginMode::Lines, 24, DrawElementsType::UnsignedByte, IntPtr::Zero);
	GL::DrawElements(BeginMode::Triangles, 36, DrawElementsType::UnsignedByte, IntPtr((int)(24 * sizeof(Byte))));

	GL::Enable(EnableCap::StencilTest);
	GL::Enable(EnableCap::CullFace);

	GL::ShadeModel(ShadingModel::Flat);

	GL::CullFace(CullFaceMode::Front);
	GL::ColorMask(false, false, false, false);
	GL::DepthMask(false);
	GL::StencilFunc(StencilFunction::Always, 1, 0x1);
	GL::StencilOp(StencilOp::Keep, StencilOp::Keep, StencilOp::Replace);
	GL::DrawElements(BeginMode::Triangles, 24, DrawElementsType::UnsignedByte, IntPtr((int)(60 * sizeof(Byte))));

	GL::CullFace(CullFaceMode::Back);
	GL::ColorMask(true, true, true, true);
	GL::DepthMask(true);
	GL::StencilFunc(StencilFunction::Notequal, 1, 0x1);
	GL::StencilOp(StencilOp::Zero, StencilOp::Zero, StencilOp::Zero);
	GL::DrawElements(BeginMode::Triangles, 24, DrawElementsType::UnsignedByte, IntPtr((int)(60 * sizeof(Byte))));

	GL::ShadeModel(ShadingModel::Smooth);

	GL::Disable(EnableCap::StencilTest);
	GL::Disable(EnableCap::CullFace);

	GL::PopMatrix();

	GL::EnableClientState(ArrayCap::TextureCoordArray);
}
