

void ClampRotation(float% val, float twopi);

public ref class IGLGeometry
{
public:
	virtual bool IsTranslucent() { return false; }

	virtual void Toggle(bool on) {}

	virtual void RenderForPicking() {}
	virtual void RenderActually() {}
};


public ref class CGLG_BMDPart : public IGLGeometry
{
public:
	CGLG_BMDPart(float scale, CBMD::CMatPolyPair^ mppair);
	virtual ~CGLG_BMDPart();

	virtual bool IsTranslucent() override { return m_Translucent; }

	virtual void Toggle(bool on) override { m_Enabled = on; }

	virtual void RenderForPicking() override { Render(3); }
	virtual void RenderActually() override { Render(2); }

private:
	array<UInt32>^ m_VBOID;
	UInt32 m_NumVertices;

	bool m_Translucent;
	UInt32 m_TextureID;
	TextureEnvMode m_TexEnvMode;

	bool m_Enabled;

	void Render(int colorvbo);
};

public ref class CGLG_ObjList : public IGLGeometry
{
public:
	CGLG_ObjList(int type, int layer, List<IPositionedObject^>^ objlist);
	virtual ~CGLG_ObjList();

	virtual bool IsTranslucent() override { return false; }

	virtual void SetLayerMask(Byte mask) { m_Enabled = (mask & (1 << m_ObjLayer)) != 0; }
	virtual void Update(int objidx);
	virtual void UpdateAll();

	virtual void RenderForPicking() override { Render(3); }
	virtual void RenderActually() override { Render(2); }

private:
	int m_ObjType;
	int m_ObjLayer;
	List<IPositionedObject^>^ m_ObjList;

	bool b_IsEmpty;
	array<UInt32>^ m_VBOID;
	int m_NumLineVertices, m_NumTriVertices;
	List<UInt32>^ m_ObjLineVtxOffsets;
	List<UInt32>^ m_ObjTriVtxOffsets;

	UInt32 m_CubeColor, m_BorderColor;

	bool m_Enabled;

	void Render(int colorvbo);
};

public ref class CGLG_ObjSelCube : public IGLGeometry
{
public:
	CGLG_ObjSelCube(UInt32 color);
	virtual ~CGLG_ObjSelCube();

	virtual bool IsTranslucent() override { return false; }

	virtual void Update(IPositionedObject^ sel);

	virtual void RenderActually() override;

private:
	IPositionedObject^ m_SelObject;
	Matrix4 m_PosMatrix;

	array<UInt32>^ m_VBOID;
};
