#include "StdAfx.h"

using namespace System;
using namespace System::Collections::Generic;
using namespace OpenTK;
using namespace OpenTK::Graphics;
using namespace OpenTK::Graphics::OpenGL;

#include "NitroROM.h"
#include "NitroFile.h"
#include "BMD.h"


void CBMD::ReadTexture(CMatPolyPair^ mppair, UInt32 texid, UInt32 palid)
{
	UInt32 texentry = m_TexChunksOffset + (texid * 20);
	UInt32 palentry = (palid == 0xFFFFFFFF) ? 0xFFFFFFFF : m_PalChunksOffset + (palid * 16);

	mppair->m_TexName = m_File->ReadString(m_File->Read32(texentry), 0);

	UInt32 texdataoffset = m_File->Read32(texentry + 0x04);
	UInt32 texdatasize = m_File->Read32(texentry + 0x08);
	UInt32 texparam = m_File->Read32(texentry + 0x10);
	mppair->m_TexParams |= texparam;

	UInt32 paldataoffset = 0xFFFFFFFF;
	UInt32 paldatasize = 0;
	if (palentry != 0xFFFFFFFF)
	{
		paldataoffset = m_File->Read32(palentry + 0x04);
		paldatasize = m_File->Read32(palentry + 0x08);
	}

	UInt32 textype = (texparam >> 26) & 0x7;
	if (textype == 0) return;

	mppair->m_TexWidth = 8 << ((texparam >> 20) & 0x7);
	mppair->m_TexHeight = 8 << ((texparam >> 23) & 0x7);

	if ((palentry == 0xFFFFFFFF) && (textype != 7))
	{
		throw gcnew Exception("BMD decoder: paletted texture with no associated palette entry; WTF");
		return;
	}

	// texture stored as 8bit ARGB
	mppair->m_Texture = gcnew array<Byte>(mppair->m_TexWidth * mppair->m_TexHeight * 4);

	switch (textype)
	{
	case 1: // a3i5
		{
			for (UInt32 in = 0, out = 0; in < texdatasize; in++, out += 4)
			{
				Byte texel = m_File->Read8(texdataoffset + in);
				UInt16 color = m_File->Read16(paldataoffset + ((texel&0x1F) << 1));

				Byte red = (color & 0x001F) << 3;
				Byte green = (color & 0x03E0) >> 2;
				Byte blue = (color & 0x7C00) >> 7;
				Byte _alpha = ((texel & 0xE0) >> 3) + ((texel & 0xE0) >> 6);
				Byte alpha = (_alpha << 3) | (_alpha >> 2);

				mppair->m_Texture[out  ] = blue;
				mppair->m_Texture[out+1] = green;
				mppair->m_Texture[out+2] = red;
				mppair->m_Texture[out+3] = alpha;
			}
		}
		break;

	case 2: // 4-color
		{
			for (int in = 0, out = 0; in < texdatasize; in++, out += 16)
			{
				Byte texels = m_File->Read8(texdataoffset + in);

				UInt16 color = m_File->Read16(paldataoffset + ((texels << 1)&0x6));
				Byte red = (color & 0x001F) << 3;
				Byte green = (color & 0x03E0) >> 2;
				Byte blue = (color & 0x7C00) >> 7;

				mppair->m_Texture[out   ] = blue;
				mppair->m_Texture[out+1 ] = green;
				mppair->m_Texture[out+2 ] = red;
				mppair->m_Texture[out+3 ] = 0xFF;

				color = m_File->Read16(paldataoffset + ((texels >> 1)&0x6));
				red = (color & 0x001F) << 3;
				green = (color & 0x03E0) >> 2;
				blue = (color & 0x7C00) >> 7;

				mppair->m_Texture[out+4 ] = blue;
				mppair->m_Texture[out+5 ] = green;
				mppair->m_Texture[out+6 ] = red;
				mppair->m_Texture[out+7 ] = 0xFF;

				color = m_File->Read16(paldataoffset + ((texels >> 3)&0x6));
				red = (color & 0x001F) << 3;
				green = (color & 0x03E0) >> 2;
				blue = (color & 0x7C00) >> 7;

				mppair->m_Texture[out+8 ] = blue;
				mppair->m_Texture[out+9 ] = green;
				mppair->m_Texture[out+10] = red;
				mppair->m_Texture[out+11] = 0xFF;

				color = m_File->Read16(paldataoffset + ((texels >> 5)&0x6));
				red = (color & 0x001F) << 3;
				green = (color & 0x03E0) >> 2;
				blue = (color & 0x7C00) >> 7;

				mppair->m_Texture[out+12] = blue;
				mppair->m_Texture[out+13] = green;
				mppair->m_Texture[out+14] = red;
				mppair->m_Texture[out+15] = 0xFF;
			}
		}
		break;

	case 3: // 16-color
		{
			for (int in = 0, out = 0; in < texdatasize; in++, out += 8)
			{
				Byte texels = m_File->Read8(texdataoffset + in);

				UInt16 color = m_File->Read16(paldataoffset + ((texels << 1)&0x1E));
				Byte red = (color & 0x001F) << 3;
				Byte green = (color & 0x03E0) >> 2;
				Byte blue = (color & 0x7C00) >> 7;

				mppair->m_Texture[out  ] = blue;
				mppair->m_Texture[out+1] = green;
				mppair->m_Texture[out+2] = red;
				mppair->m_Texture[out+3] = 0xFF;

				color = m_File->Read16(paldataoffset + ((texels >> 3)&0x1E));
				red = (color & 0x001F) << 3;
				green = (color & 0x03E0) >> 2;
				blue = (color & 0x7C00) >> 7;

				mppair->m_Texture[out+4] = blue;
				mppair->m_Texture[out+5] = green;
				mppair->m_Texture[out+6] = red;
				mppair->m_Texture[out+7] = 0xFF;
			}
		}
		break;

	case 4: // 256-color
		{
			for (int in = 0, out = 0; in < texdatasize; in++, out += 4)
			{
				Byte texel = m_File->Read8(texdataoffset + in);

				UInt16 color = m_File->Read16(paldataoffset + (texel << 1));
				Byte red = (color & 0x001F) << 3;
				Byte green = (color & 0x03E0) >> 2;
				Byte blue = (color & 0x7C00) >> 7;

				mppair->m_Texture[out  ] = blue;
				mppair->m_Texture[out+1] = green;
				mppair->m_Texture[out+2] = red;
				mppair->m_Texture[out+3] = 0xFF;
			}
		}
		break;

	case 5: // motherfucking compressed
		{   // as usual, shitty code but works
			int yout = 0, xout = 0;

			for (int in = 0; in < texdatasize; in += 4)
			{
				UInt32 blox = m_File->Read32(texdataoffset + in);
				UInt16 palidx_data = m_File->Read16(texdataoffset + texdatasize + (in>>1));

				for (int y = 0; y < 4; y++)
				{
					for (int x = 0; x < 4; x++)
					{
						Byte texel = blox & 0x3;
						blox >>= 2;

						UInt32 pal_offset = (palidx_data & 0x3FFF) << 2;
						UInt16 color_mode = palidx_data >> 14;
						UInt32 color;

						switch (texel)
						{
						case 0: color = m_File->Read16(paldataoffset + pal_offset); break;
						case 1: color = m_File->Read16(paldataoffset + pal_offset + 2); break;
						case 2:
							{
								switch (color_mode)
								{
								case 0:
								case 2: color = m_File->Read16(paldataoffset + pal_offset + 4); break;
								case 1:
									{
										UInt16 c0 = m_File->Read16(paldataoffset + pal_offset);
										UInt16 c1 = m_File->Read16(paldataoffset + pal_offset + 2);

										UInt16 red = ((c0 & 0x001F) + (c1 & 0x001F)) >> 1;
										UInt16 green = ((c0 & 0x03E0) + (c1 & 0x03E0)) >> 1;
										UInt16 blue = ((c0 & 0x7C00) + (c1 & 0x7C00)) >> 1;
										color = (red & 0x001F) | (green & 0x03E0) | (blue & 0x7C00);
									}
									break;
								case 3:
									{
										UInt16 c0 = m_File->Read16(paldataoffset + pal_offset);
										UInt16 c1 = m_File->Read16(paldataoffset + pal_offset + 2);

										UInt16 red = ((c0 & 0x001F)*5 + (c1 & 0x001F)*3) >> 3;
										UInt16 green = ((c0 & 0x03E0)*5 + (c1 & 0x03E0)*3) >> 3;
										UInt16 blue = ((c0 & 0x7C00)*5 + (c1 & 0x7C00)*3) >> 3;
										color = (red & 0x001F) | (green & 0x03E0) | (blue & 0x7C00);
									}
									break;
								}
							}
							break;
						case 3:
							{
								switch (color_mode)
								{
								case 0:
								case 1: color = 0xFFFFFFFF; break;
								case 2: color = m_File->Read16(paldataoffset + pal_offset + 6); break;
								case 3:
									{
										UInt16 c0 = m_File->Read16(paldataoffset + pal_offset);
										UInt16 c1 = m_File->Read16(paldataoffset + pal_offset + 2);

										UInt16 red = ((c0 & 0x001F)*3 + (c1 & 0x001F)*5) >> 3;
										UInt16 green = ((c0 & 0x03E0)*3 + (c1 & 0x03E0)*5) >> 3;
										UInt16 blue = ((c0 & 0x7C00)*3 + (c1 & 0x7C00)*5) >> 3;
										color = (red & 0x001F) | (green & 0x03E0) | (blue & 0x7C00);
									}
									break;
								}
							}
							break;
						}

						int out = ((yout * mppair->m_TexWidth) + xout) * 4;
						int yoff = y * mppair->m_TexWidth * 4;
						int xoff = x * 4;

						if (color == 0xFFFFFFFF)
						{
							mppair->m_Texture[out+yoff+xoff  ] = 0;
							mppair->m_Texture[out+yoff+xoff+1] = 0;
							mppair->m_Texture[out+yoff+xoff+2] = 0;
							mppair->m_Texture[out+yoff+xoff+3] = 0;
						}
						else
						{
							Byte red = (color & 0x001F) << 3;
							Byte green = (color & 0x03E0) >> 2;
							Byte blue = (color & 0x7C00) >> 7;

							mppair->m_Texture[out+yoff+xoff  ] = blue;
							mppair->m_Texture[out+yoff+xoff+1] = green;
							mppair->m_Texture[out+yoff+xoff+2] = red;
							mppair->m_Texture[out+yoff+xoff+3] = 0xFF;
						}
					}
				}

				xout += 4;
				if (xout >= mppair->m_TexWidth)
				{
					xout = 0;
					yout += 4;
				}
			}
		}
		break;

	case 6: // a5i3
		{
			for (int in = 0, out = 0; in < texdatasize; in++, out += 4)
			{
				Byte texel = m_File->Read8(texdataoffset + in);
				UInt16 color = m_File->Read16(paldataoffset + ((texel&0x07) << 1));

				Byte red = (color & 0x001F) << 3;
				Byte green = (color & 0x03E0) >> 2;
				Byte blue = (color & 0x7C00) >> 7;
				Byte alpha = (texel & 0xF8) | ((texel & 0xF8) >> 5);

				mppair->m_Texture[out  ] = blue;
				mppair->m_Texture[out+1] = green;
				mppair->m_Texture[out+2] = red;
				mppair->m_Texture[out+3] = alpha;
			}
		}
		break;

	case 7: // direct color
		{
			for (int in = 0, out = 0; in < texdatasize; in += 2, out += 4)
			{
				UInt16 color = m_File->Read16(texdataoffset + in);
				Byte red = (color & 0x001F) << 3;
				Byte green = (color & 0x03E0) >> 2;
				Byte blue = (color & 0x7C00) >> 7;

				mppair->m_Texture[out  ] = blue;
				mppair->m_Texture[out+1] = green;
				mppair->m_Texture[out+2] = red;
				mppair->m_Texture[out+3] = 0xFF;
			}
		}
		break;
	}
}

void CBMD::ProcessGXCommand(CMatPolyPair^ mppair, Byte cmd, UInt32% pos)
{
	CGeometryBlock^ gblock = nullptr;
	if (mppair->m_Geometry->Count > 0)
		gblock = mppair->m_Geometry[mppair->m_Geometry->Count-1];

	//if ((cmd&0xF0)==0x10 && (cmd!=0x14))
	//	throw gcnew Exception(String::Format("MATRIX COMMAND {0:X2}", cmd));

	switch (cmd)
	{
	// nop
	case 0x00: break;

	// matrix commands
	case 0x10: pos += 4; break;
	case 0x11: break;
	case 0x12: pos += 4; break;
	case 0x13: pos += 4; break;
	case 0x14: pos += 4; break;
	case 0x15: break;
	case 0x16: pos += 64; break;
	case 0x17: pos += 48; break;
	case 0x18: pos += 64; break;
	case 0x19: pos += 48; break;
	case 0x1A: pos += 36; break;
	case 0x1B: pos += 12; break;
	case 0x1C: pos += 12; break;
	
	// set color
	case 0x20:
		{
			UInt32 raw = m_File->Read32(pos); pos += 4;

			Byte red = (raw << 3) & 0xFC;
			Byte green = (raw >> 2) & 0xFC;
			Byte blue = (raw >> 7) & 0xFC;
			m_CurVertex.m_Color = red | (green<<8) | (blue<<16);
		}
		break;

	// normal
	case 0x21: pos += 4; break;

	// texcoord
	case 0x22:
		{
			UInt32 param = m_File->Read32(pos); pos += 4;

			Int16 s = (Int16)(param & 0xFFFF);
			Int16 t = (Int16)(param >> 16);
			m_CurVertex.m_Texcoord.X = (float)s / (16.f * (float)mppair->m_TexWidth);
			m_CurVertex.m_Texcoord.Y = (float)t / (16.f * (float)mppair->m_TexHeight);
		}
		break;

	// define vertex
	case 0x23:
		{
			UInt32 param1 = m_File->Read32(pos); pos += 4;
			UInt32 param2 = m_File->Read32(pos); pos += 4;

			Int16 x = (Int16)(param1 & 0xFFFF);
			Int16 y = (Int16)(param1 >> 16);
			Int16 z = (Int16)(param2 & 0xFFFF);
			m_CurVertex.m_Position.X = (float)x / 4096.f;
			m_CurVertex.m_Position.Y = (float)y / 4096.f;
			m_CurVertex.m_Position.Z = (float)z / 4096.f;

			gblock->m_VertexList->Add(m_CurVertex);
		}
		break;

	case 0x24:
		{
			UInt32 param = m_File->Read32(pos); pos += 4;

			Int16 x = (Int16)((param << 6) & 0xFFC0);
			Int16 y = (Int16)((param >> 4) & 0xFFC0);
			Int16 z = (Int16)((param >> 14) & 0xFFC0);
			m_CurVertex.m_Position.X = (float)x / 4096.f;
			m_CurVertex.m_Position.Y = (float)y / 4096.f;
			m_CurVertex.m_Position.Z = (float)z / 4096.f;

			gblock->m_VertexList->Add(m_CurVertex);
		}
		break;

	case 0x25:
		{
			UInt32 param = m_File->Read32(pos); pos += 4;

			Int16 x = (Int16)(param & 0xFFFF);
			Int16 y = (Int16)(param >> 16);
			m_CurVertex.m_Position.X = (float)x / 4096.f;
			m_CurVertex.m_Position.Y = (float)y / 4096.f;

			gblock->m_VertexList->Add(m_CurVertex);
		}
		break;

	case 0x26:
		{
			UInt32 param = m_File->Read32(pos); pos += 4;

			Int16 x = (Int16)(param & 0xFFFF);
			Int16 z = (Int16)(param >> 16);
			m_CurVertex.m_Position.X = (float)x / 4096.f;
			m_CurVertex.m_Position.Z = (float)z / 4096.f;

			gblock->m_VertexList->Add(m_CurVertex);
		}
		break;

	case 0x27:
		{
			UInt32 param = m_File->Read32(pos); pos += 4;

			Int16 y = (Int16)(param & 0xFFFF);
			Int16 z = (Int16)(param >> 16);
			m_CurVertex.m_Position.Y = (float)y / 4096.f;
			m_CurVertex.m_Position.Z = (float)z / 4096.f;

			gblock->m_VertexList->Add(m_CurVertex);
		}
		break;

	case 0x28:
		{
			UInt32 param = m_File->Read32(pos); pos += 4;

			Int16 x = (Int16)((param << 6) & 0xFFC0);
			Int16 y = (Int16)((param >> 4) & 0xFFC0);
			Int16 z = (Int16)((param >> 14) & 0xFFC0);
			m_CurVertex.m_Position.X += ((float)x / 262144.f);
			m_CurVertex.m_Position.Y += ((float)y / 262144.f);
			m_CurVertex.m_Position.Z += ((float)z / 262144.f);

			gblock->m_VertexList->Add(m_CurVertex);
		}
		break;

	case 0x29: pos += 4; break;
	case 0x2A: pos += 4; break;
	case 0x2B: pos += 4; break;

	// lighting commands
	case 0x30: pos += 4; break;
	case 0x31: pos += 4; break;
	case 0x32: pos += 4; break;
	case 0x33: pos += 4; break;
	case 0x34: pos += 128; break;

	// Begin vertex list
	case 0x40: 
		{
			UInt32 param = m_File->Read32(pos); pos += 4;

			gblock = gcnew CGeometryBlock();
			mppair->m_Geometry->Add(gblock);
			gblock->m_Type = param & 0x00000003;
			gblock->m_VertexList = gcnew List<SVertex>();
		}
		break;

	// End vertex list (dummy)
	case 0x41: break;

	// misc.
	case 0x50: pos += 4; break;
	case 0x60: pos += 4; break;
	case 0x70: pos += 12; break;
	case 0x71: pos += 8; break;
	case 0x72: pos += 4; break;

	default: throw gcnew Exception(String::Format("Unknown GX command {0:X2} in BMD file", cmd)); return;
	}
}

CBMD::CBMD(CNitroFile^ file)
{
	m_File = file;

	m_ScaleFactor = 1 << m_File->Read32(0x0);

	m_NumModelChunks = m_File->Read32(0x04);
	m_ModelChunksOffset = m_File->Read32(0x08);

	m_NumPolyChunks = m_File->Read32(0x0C);
	m_PolyChunksOffset = m_File->Read32(0x10);
	m_NumTexChunks = m_File->Read32(0x14);
	m_TexChunksOffset = m_File->Read32(0x18);
	m_NumPalChunks = m_File->Read32(0x1C);
	m_PalChunksOffset = m_File->Read32(0x20);
	m_NumMatChunks = m_File->Read32(0x24);
	m_MatChunksOffset = m_File->Read32(0x28);

	m_ModelChunks = gcnew array<CModelChunk^>(m_NumModelChunks);

	for (UInt32 c = 0; c < m_NumModelChunks; c++)
	{
		CModelChunk^ mdchunk = gcnew CModelChunk();
		m_ModelChunks[c] = mdchunk;

		UInt32 mdchunkoffset = m_ModelChunksOffset + (c * 64);

		mdchunk->m_ID = m_File->Read32(mdchunkoffset);
		mdchunk->m_Name = m_File->ReadString(m_File->Read32(mdchunkoffset + 0x04), 0);

		UInt32 numpairs = m_File->Read32(mdchunkoffset + 0x30);
		UInt32 matlist = m_File->Read32(mdchunkoffset + 0x34);
		UInt32 polylist = m_File->Read32(mdchunkoffset + 0x38);

		mdchunk->m_MatPolyPairs = gcnew array<CMatPolyPair^>(numpairs);

		for (UInt32 i = 0; i < numpairs; i++)
		{
			CMatPolyPair^ mppair = gcnew CMatPolyPair();
			mdchunk->m_MatPolyPairs[i] = mppair;

			Byte matID = m_File->Read8(matlist+i);
			Byte polyID = m_File->Read8(polylist+i);

			UInt32 mchunkoffset = m_MatChunksOffset + (matID * 48);

			mppair->m_ID = i;
			mppair->m_Name = m_File->ReadString(m_File->Read32(mchunkoffset), 0);
			UInt32 texid = m_File->Read32(mchunkoffset + 0x04);
			UInt32 palid = m_File->Read32(mchunkoffset + 0x08);
			mppair->m_TexParams = m_File->Read32(mchunkoffset + 0x20);
			mppair->m_PolyAttributes = m_File->Read32(mchunkoffset + 0x24);
			mppair->m_DifAmbColors = m_File->Read32(mchunkoffset + 0x28);
			mppair->m_SpeEmiColors = m_File->Read32(mchunkoffset + 0x2C);

			if (texid != 0xFFFFFFFF)
				ReadTexture(mppair, texid, palid);
			else
				mppair->m_Texture = nullptr;

			UInt32 pchunkoffset = m_File->Read32(m_PolyChunksOffset + (polyID*8) + 4);
			UInt32 dloffset = m_File->Read32(pchunkoffset + 0x0C);
			UInt32 dlsize = m_File->Read32(pchunkoffset + 0x08);

			mppair->m_Geometry = gcnew List<CGeometryBlock^>();

			m_CurVertex.m_Position = Vector3(0,0,0);
			m_CurVertex.m_Texcoord = Vector2(0,0);
			if (mppair->m_DifAmbColors & 0x8000)
			{
				Byte red = (mppair->m_DifAmbColors << 3) & 0xFC;
				Byte green = (mppair->m_DifAmbColors >> 2) & 0xFC;
				Byte blue = (mppair->m_DifAmbColors >> 7) & 0xFC;
				m_CurVertex.m_Color = red | (green<<8) | (blue<<16);
			}
			else
				m_CurVertex.m_Color = 0x00000000;

			UInt32 dlend = dloffset+dlsize;
			for (UInt32 pos = dloffset; pos < dlend; )
			{
				Byte cmd1 = m_File->Read8(pos++);
				Byte cmd2 = m_File->Read8(pos++);
				Byte cmd3 = m_File->Read8(pos++);
				Byte cmd4 = m_File->Read8(pos++);

				ProcessGXCommand(mppair, cmd1, pos);
				ProcessGXCommand(mppair, cmd2, pos);
				ProcessGXCommand(mppair, cmd3, pos);
				ProcessGXCommand(mppair, cmd4, pos);
			}
		}
	}
}

CBMD::~CBMD()
{
	for each (CModelChunk^ mdchunk in m_ModelChunks)
		delete mdchunk;

	delete m_File;
}
