#include "StdAfx.h"
#include "EditorMainForm.h"

namespace sm64ds_editor
{

void EditorMainForm::ReadObjTable(UInt32 offset)
{
	if (offset == 0) return;

	offset -= m_BaseRAMAddr;
	UInt32 num_subtbl = m_LevelOverlay->Read32(offset);
	UInt32 subtbl_offset = m_LevelOverlay->Read32(offset + 4) - m_BaseRAMAddr;

	for (UInt32 i = 0; i < num_subtbl; i++)
	{
		UInt32 cur_offset = subtbl_offset + (i*8);
		Byte flags = m_LevelOverlay->Read8(cur_offset);
		Byte numentries = m_LevelOverlay->Read8(cur_offset+1);
		UInt32 entries_offset = m_LevelOverlay->Read32(cur_offset+4) - m_BaseRAMAddr;

		Byte tbltype = flags & 0x1F;
		Byte layer = flags >> 5;

		switch (tbltype)
		{
		case 0: 
			for (Byte j = 0; j < numentries; j++)
				m_StdObjects[layer]->Add(gcnew CStdObject(m_LevelOverlay, entries_offset + (j*16), layer, m_StdObjects[layer]->Count, false));
			break;

#if 0
		case 1:
			for (Byte j = 0; j < numentries; j++)
				m_ExtraObjects[layer]->Add(gcnew CExtraObject(m_LevelOverlay, entries_offset + (j*16), layer, m_ExtraObjects[layer]->Count));
			break;
#endif

		case 5: 
			for (Byte j = 0; j < numentries; j++)
				m_SimpleObjects[layer]->Add(gcnew CSimpleObject(m_LevelOverlay, entries_offset + (j*8), layer, m_SimpleObjects[layer]->Count, false));
			break;

		default: break;
		}
	}
}

void EditorMainForm::LoadLevelData()
{
	m_StdObjects = gcnew TPosObjList(8);
	m_SimpleObjects = gcnew TPosObjList(8);
	m_ExtraObjects = gcnew TPosObjList(8);
	for (int i = 0; i < 8; i++) 
	{
		m_StdObjects[i] = gcnew List<IPositionedObject^>();
		m_SimpleObjects[i] = gcnew List<IPositionedObject^>();
		m_ExtraObjects[i] = gcnew List<IPositionedObject^>();
	}

	this->Text = String::Format("[{0}] {1} - SM64DS Editor v1.0", m_LevelID, gcnew String(g_LevelNames[m_LevelID]));

	m_ROM->BeginRW();

	m_LevelOverlay = gcnew CNitroOverlay(m_ROM, 8+m_LevelID);
	m_BaseRAMAddr = m_LevelOverlay->GetRAMAddr();
	m_LevelOffset = m_ROM->GetLevelAddress(m_LevelID) - m_BaseRAMAddr;

	m_LevelSettings = gcnew CLevelSettings(m_LevelID, m_LevelOverlay, m_LevelOffset);
	m_MyBackground = m_LevelSettings->Background;

	ReadObjTable(m_LevelOverlay->Read32(m_LevelOffset + 0x04));

	UInt32 ObjTablesList = m_LevelOverlay->Read32(m_LevelOffset + 0x10) - m_BaseRAMAddr;
	Byte NumObjTables = m_LevelOverlay->Read8(m_LevelOffset + 0x14);

	for (Byte t = 0; t < NumObjTables; t++)
		ReadObjTable(m_LevelOverlay->Read32(ObjTablesList + (t*12)));

	m_LayerMask = 0x03;
	btnLayer0->Checked = true;
	btnLayer1->Checked = true;
	PopulateObjList(0);

	if (m_LevelSettings->Background)
	{
		try
		{
			m_SkyboxModel = gcnew CBMD(gcnew CNitroFile(m_ROM, m_ROM->GetFileIDFromName(String::Format("data/vrbox/vr{0:D2}.bmd", m_LevelSettings->Background))));
		}
		catch (Exception^)
		{
			sblStatusLabel->Text = String::Format("Failed to load background 'data/vrbox/vr{0:D2}.bmd'", m_LevelSettings->Background);
			m_SkyboxModel = nullptr;
		}
	}
	else
		m_SkyboxModel = nullptr;

	UInt16 bmd_fileid = m_LevelOverlay->Read16(m_LevelOffset + 8);
	if (bmd_fileid >= 0x8000) { throw gcnew Exception("Loading level files from NARC archives not supported yet."); return; }
	m_LevelModel = gcnew CBMD(gcnew CNitroFile(m_ROM, m_ROM->GetFileIDFromInternalID(bmd_fileid)));

#if 0
	kcltest = gcnew List<Vector3>();
	CNitroFile^ kclfile = gcnew CNitroFile(m_ROM, 1510/*1332*/);
	UInt32 start = kclfile->Read32(0x00);
	UInt32 end = kclfile->Read32(0x04);
	UInt32 start2 = kclfile->Read32(0x08) + 0x14;
	UInt32 end2 = kclfile->Read32(0x0C);
	//for (UInt32 lol = start2; lol < end2; lol += 16)
	for (UInt32 lol = start; lol < end; lol += 12)
	{
		//UInt16 pt = kclfile->Read16(lol+0); // 0 is the nicest one! :D
		UInt16 pt = (lol-start)/12;
	//	UInt16 pt2 = kclfile->Read16(lol+6);
	//	UInt16 pt3 = kclfile->Read16(lol+8);

	//	array<UInt16>^ pts = {pt1, pt2, pt3};
	//	for each (UInt16 pt in pts)
		{
			//if (pt>0xF5) continue;
			if ((start+(pt*12)) >= end) 
			{
			//	MessageBox::Show(String::Format(" | plane {0} (pt {1}) out of range ({2} points max)", (lol-start2)/16, pt, (end-start)/12));
				continue;
			}
			//Int16 x = (Int16)kclfile->Read16(lol);
			//Int16 y = (Int16)kclfile->Read16(lol+2);
			//Int16 z = (Int16)kclfile->Read16(lol+4);
			UInt32 _x = kclfile->Read32(start+(pt*12));
			UInt32 _y = kclfile->Read32(start+(pt*12)+4);
			UInt32 _z = kclfile->Read32(start+(pt*12)+8);

			//if (_x & 0x00800000) _x |= 0xFF000000;
			//if (_y & 0x00800000) _y |= 0xFF000000;
			//if (_z & 0x00800000) _z |= 0xFF000000;

			Int32 x = (Int32)_x;
			Int32 y = (Int32)_y;
			Int32 z = (Int32)_z;

			float factor = 64000.f;
			kcltest->Add(Vector3((float)x/factor, (float)y/factor, (float)z/factor));
			//kcltest->Add(Vector3((float)x, (float)y, (float)z));
		}
	}
	delete kclfile;
#endif

	m_ROM->EndRW();
}

void EditorMainForm::DeleteLevelData()
{
	for each (IGLGeometry^ glg in m_OSkyboxGeometry)
		delete glg;
	for each (IGLGeometry^ glg in m_TSkyboxGeometry)
		delete glg;
	for each (IGLGeometry^ glg in m_OLevelGeometry)
		delete glg;
	for each (IGLGeometry^ glg in m_TLevelGeometry)
		delete glg;
	for each (KeyValuePair<UInt32,CGLG_ObjList^>^ ol in m_ObjCubesList)
		delete ol->Value;

	delete m_OSkyboxGeometry;
	delete m_TSkyboxGeometry;
	delete m_OLevelGeometry;
	delete m_TLevelGeometry;
	delete m_ObjCubesList;
	delete m_SelCube;
	delete m_HoverCube;

	delete m_SkyboxModel;
	delete m_LevelModel;

	for (int i = 0; i < 8; i++)
	{
		for each (IPositionedObject^ obj in m_StdObjects[i])
			delete obj;
		for each (IPositionedObject^ obj in m_SimpleObjects[i])
			delete obj;

		delete m_StdObjects[i];
		delete m_SimpleObjects[i];
		delete m_ExtraObjects[i];
	}

	delete m_StdObjects;
	delete m_SimpleObjects;
	delete m_ExtraObjects;

	delete m_LevelSettings;
	delete m_LevelOverlay;
}

void EditorMainForm::FixObjectTable(UInt32 offset, UInt32 fixstart, int delta)
{
	if (offset == 0) return;

	UInt32 temp;
#define FIX(offset) \
	temp = m_LevelOverlay->Read32(offset); \
	if (temp >= fixstart) \
	{ \
		temp += delta; \
		m_LevelOverlay->Write32(offset, temp); \
	}

	offset -= m_BaseRAMAddr;

	FIX(offset + 4);

	UInt32 num_subtbl = m_LevelOverlay->Read32(offset);
	UInt32 subtbl_offset = m_LevelOverlay->Read32(offset + 4) - m_BaseRAMAddr;

	for (UInt32 i = 0; i < num_subtbl; i++)
	{
		UInt32 cur_offset = subtbl_offset + (i*8);
		FIX(cur_offset + 4);
	}

#undef FIX
}

void EditorMainForm::AddObject(int type, int layer)
{
	MessageBox::Show("You are truly extraordinary to have managed to reach this place.\nHowever, this feature doesn't work, and your level is now most likely corrupted. Enjoy!", 
		"SM64DS Editor - oops", MessageBoxButtons::OK, MessageBoxIcon::Information);

	if (type < 0) return;
	array<Byte>^ tbltypes = {0, 5};
	array<UInt32>^ tblentrysizes = {16, 8};

	// find a decent place for the object
	UInt32 place = 0xFFFFFFFF;
	UInt32 ObjTablesList = m_LevelOverlay->Read32(m_LevelOffset + 0x10) - m_BaseRAMAddr;
	Byte NumObjTables = m_LevelOverlay->Read8(m_LevelOffset + 0x14);

	for (Byte t = 0; t < NumObjTables; t++)
	{
		UInt32 offset = m_LevelOverlay->Read32(ObjTablesList + (t*12));
		if (offset == 0) continue;

		offset -= m_BaseRAMAddr;
		UInt32 num_subtbl = m_LevelOverlay->Read32(offset);
		UInt32 subtbl_offset = m_LevelOverlay->Read32(offset + 4) - m_BaseRAMAddr;

		for (UInt32 i = 0; i < num_subtbl; i++)
		{
			UInt32 cur_offset = subtbl_offset + (i*8);

			Byte flags = m_LevelOverlay->Read8(cur_offset);
			Byte numentries = m_LevelOverlay->Read8(cur_offset+1);
			UInt32 entries_offset = m_LevelOverlay->Read32(cur_offset+4) - m_BaseRAMAddr;

			Byte this_tbltype = flags & 0x1F;
			Byte this_layer = flags >> 5;

			if (this_tbltype != tbltypes[type]) continue;
			if (this_layer != layer) continue;
			if (numentries == 0xFF) continue;

			place = entries_offset + (tblentrysizes[type] * numentries);
			m_LevelOverlay->Write8(cur_offset+1, numentries + 1);
		}

		if (place == 0xFFFFFFFF)
		{
			// TODO!
			throw gcnew Exception("Failed to add object: no available space found.\nIf you see this, it means Mega-Mario should get off his lazy fat ass.");
			return;
		}

		break;
	}

	UInt32 delta = tblentrysizes[type];

	// move data next to that place
	m_LevelOverlay->WriteBlock(place + delta, m_LevelOverlay->ReadBlock(place, m_LevelOverlay->m_Data->Length - place));

	// fix offsets'n'stuff
	UInt32 fixstart = m_BaseRAMAddr + place;
	UInt32 fixend = m_BaseRAMAddr + m_LevelOverlay->m_Data->Length - delta;
	UInt32 temp;
#define FIX(offset) \
	temp = m_LevelOverlay->Read32(offset); \
	if ((temp >= fixstart) && (temp < fixend)) \
	{ \
		temp += delta; \
		m_LevelOverlay->Write32(offset, temp); \
	}

	FIX(m_LevelOffset);
	FIX(m_LevelOffset + 0x04);
	FIX(m_LevelOffset + 0x10);

	FixObjectTable(m_LevelOverlay->Read32(m_LevelOffset + 0x04), fixstart, delta);

	ObjTablesList = m_LevelOverlay->Read32(m_LevelOffset + 0x10) - m_BaseRAMAddr;
	for (Byte t = 0; t < NumObjTables; t++)
	{
		FIX(ObjTablesList + (t*12));
		FixObjectTable(m_LevelOverlay->Read32(ObjTablesList + (t*12)), fixstart, delta);
	}

	array<UInt32>^ tofix = m_LevelOverlay->m_ROM->FixTables(m_LevelID, fixstart, delta);
	for each (UInt32 actorptr in tofix)
	{
		actorptr -= m_BaseRAMAddr;

		FIX(actorptr);
		for (UInt32 i = 0x20; i < 0xA4; i += 4)
		{
			FIX(actorptr + i);
		}
	}

#undef FIX

	// create the new object
	IPositionedObject^ obj;
	OBJLISTS;

	switch (type)
	{
	case 0: obj = gcnew CStdObject(m_LevelOverlay, place, layer, m_StdObjects[layer]->Count, true); break;
	case 1: obj = gcnew CSimpleObject(m_LevelOverlay, place, layer, m_SimpleObjects[layer]->Count, true); break;
	}

	objlists[m_CurObjType][layer]->Add(obj);

	TreeNode^ catnode = tvObjectList->Nodes->Find(String::Format("cat_{0:X8}", (m_CurObjType<<28) | (layer<<24)), true)[0];
	TreeNode^ objnode = catnode->Nodes->Add(String::Format("{0:X8}", obj->m_UniqueID), obj->GetDesc());
	objnode->Tag = obj;
	m_ObjCubesList[(m_CurObjType<<28) | (layer<<24)]->UpdateAll();

	m_MouseDown = 1;
	m_Selection = (m_CurObjType<<28) | (layer<<24) | obj->m_UniqueID;
	m_UnderCursor = m_Selection;
	m_SelectedObj = obj;
	m_SelCube->Update(obj);
	SetupObjEditPanel(m_CurObjType);
	obj->FillEditPanel(tlpObjEditPanel, 0x0);

	tvObjectList->SelectedNode = objnode;
	tvObjectList->SelectedNode->EnsureVisible();
}

void EditorMainForm::PopulateObjList(int type)
{
	if (type < 0) return;

	array<String^>^ layerstr = {"All stars", "Star 1", "Star 2", "Star 3", "Star 4", "Star 5", "Star 6", "Star 7"};
	OBJLISTS;

	tvObjectList->Nodes->Clear();
	for (int i = 0; i < 8; i++)
	{
		if (!(m_LayerMask & (1 << i))) continue;

		TreeNode^ curlayer = tvObjectList->Nodes->Add(String::Format("cat_{0:X8}", ((type<<28) | (i<<24))), layerstr[i]);
		curlayer->Tag = UInt32(i);

		for each (IPositionedObject^ obj in objlists[type][i])
		{
			TreeNode^ node = curlayer->Nodes->Add(String::Format("{0:X8}", obj->m_UniqueID), obj->GetDesc());
			node->Tag = obj;
		}

		curlayer->Expand();
	}

	if (m_LayerMask)
		tvObjectList->Nodes[0]->EnsureVisible();

	m_Selection = 0xFFFFFFFF;
	if (m_SelCube) m_SelCube->Update(nullptr);
}

void EditorMainForm::SetupObjEditPanel(int type)
{
#define ADD_OBJLIST(y) \
	tlpObjEditPanel->Controls->Add(gcnew Label(), 0, y); \
	tlpObjEditPanel->Controls[y*2]->Text = "Object ID:"; \
	tlpObjEditPanel->Controls->Add(gcnew ComboBox(), 1, y); \
	tlpObjEditPanel->Controls[y*2 + 1]->Dock = DockStyle::Fill; \
	for (int n = 0; n < 512; n++) \
		((ComboBox^)tlpObjEditPanel->Controls[y*2 + 1])->Items->Add(String::Format("{0:X4} - {1}", n, m_ROM->GetObjectName(n))); \
	((ComboBox^)tlpObjEditPanel->Controls[y*2 + 1])->SelectedIndexChanged += gcnew System::EventHandler(this, &EditorMainForm::objParamChanged);

#define ADD_UPDOWN(y, txt, dp, min, max, step) \
	tlpObjEditPanel->Controls->Add(gcnew Label(), 0, y); \
	tlpObjEditPanel->Controls[y*2]->Text = txt; \
	tlpObjEditPanel->Controls->Add(gcnew NumericUpDown(), 1, y); \
	tlpObjEditPanel->Controls[y*2 + 1]->Dock = DockStyle::Fill; \
	((NumericUpDown^)tlpObjEditPanel->Controls[y*2 + 1])->DecimalPlaces = dp; \
	((NumericUpDown^)tlpObjEditPanel->Controls[y*2 + 1])->Increment = Decimal(step); \
	((NumericUpDown^)tlpObjEditPanel->Controls[y*2 + 1])->Minimum = Decimal(min); \
	((NumericUpDown^)tlpObjEditPanel->Controls[y*2 + 1])->Maximum = Decimal(max); \
	((NumericUpDown^)tlpObjEditPanel->Controls[y*2 + 1])->ValueChanged += gcnew System::EventHandler(this, &EditorMainForm::objParamChanged);

#define ADD_TEXTBOX(y, txt) \
	tlpObjEditPanel->Controls->Add(gcnew Label(), 0, y); \
	tlpObjEditPanel->Controls[y*2]->Text = txt; \
	tlpObjEditPanel->Controls->Add(gcnew TextBox(), 1, y); \
	tlpObjEditPanel->Controls[y*2 + 1]->Dock = DockStyle::Fill; \
	tlpObjEditPanel->Controls[y*2 + 1]->TextChanged += gcnew System::EventHandler(this, &EditorMainForm::objParamChanged);

	if (m_SelObjType != type)
	{
		tlpObjEditPanel->Controls->Clear();

		switch (type)
		{
		case 0:
			 tlpObjEditPanel->RowCount = 8;
			 for each (RowStyle^ rs in tlpObjEditPanel->RowStyles)
				 rs->SizeType = SizeType::AutoSize;

			 ADD_OBJLIST(0);
			 ADD_UPDOWN(1, "X position:", 3, -0x8000/1000.f, 0x7FFF/1000.f, 1.f/1000.f);
			 ADD_UPDOWN(2, "Y position:", 3, -0x8000/1000.f, 0x7FFF/1000.f, 1.f/1000.f);
			 ADD_UPDOWN(3, "Z position:", 3, -0x8000/1000.f, 0x7FFF/1000.f, 1.f/1000.f);
			 ADD_UPDOWN(4, "Y rotation:", 9, -0x8000*22.5f/4096.f, 0x7FFF*22.5f/4096.f, 22.5f/4096.f);
			 ADD_TEXTBOX(5, "Parameter 1:");
			 ADD_TEXTBOX(6, "Parameter 2:");
			 ADD_TEXTBOX(7, "Parameter 3:");
			 break;

		case 1:
			tlpObjEditPanel->RowCount = 6;
			 for each (RowStyle^ rs in tlpObjEditPanel->RowStyles)
				 rs->SizeType = SizeType::AutoSize;

			 ADD_OBJLIST(0);
			 ADD_UPDOWN(1, "X position:", 3, -0x8000/1000.f, 0x7FFF/1000.f, 1.f/1000.f);
			 ADD_UPDOWN(2, "Y position:", 3, -0x8000/1000.f, 0x7FFF/1000.f, 1.f/1000.f);
			 ADD_UPDOWN(3, "Z position:", 3, -0x8000/1000.f, 0x7FFF/1000.f, 1.f/1000.f);
			 ADD_TEXTBOX(4, "Parameter 1:");
			 ADD_TEXTBOX(5, "Parameter 2:");
			 break;

		case 2:
			tlpObjEditPanel->RowCount = 8;
			 for each (RowStyle^ rs in tlpObjEditPanel->RowStyles)
				 rs->SizeType = SizeType::AutoSize;

			 ADD_OBJLIST(0);
			 ADD_UPDOWN(1, "X position:", 3, -0x8000/1000.f, 0x7FFF/1000.f, 1.f/1000.f);
			 ADD_UPDOWN(2, "Y position:", 3, -0x8000/1000.f, 0x7FFF/1000.f, 1.f/1000.f);
			 ADD_UPDOWN(3, "Z position:", 3, -0x8000/1000.f, 0x7FFF/1000.f, 1.f/1000.f);
			 ADD_UPDOWN(4, "Y rotation:", 9, -0x8000*22.5f/4096.f, 0x7FFF*22.5f/4096.f, 22.5f/4096.f);
			 ADD_TEXTBOX(5, "Parameter 1:");
			 ADD_TEXTBOX(6, "Parameter 2:");
			 ADD_TEXTBOX(7, "Parameter 3:");
			 break;
		}

		m_SelObjType = type;
	}

#undef ADD_OBJLIST
#undef ADD_UPDOWN
#undef ADD_TEXTBOX
}


void EditorMainForm::BuildSkyboxGeometry()
{
	for each (IGLGeometry^ glg in m_OSkyboxGeometry)
		delete glg;
	for each (IGLGeometry^ glg in m_TSkyboxGeometry)
		delete glg;

	m_OSkyboxGeometry->Clear();
	m_TSkyboxGeometry->Clear();

	if (!m_SkyboxModel) return;

	int nchunks = m_SkyboxModel->m_ModelChunks->Length;

	float scale = (float)m_SkyboxModel->m_ScaleFactor / 75.f;

	for (int c = 0; c < nchunks; c++)
	{
		CBMD::CModelChunk^ mdchunk = m_SkyboxModel->m_ModelChunks[c];
		int npairs = mdchunk->m_MatPolyPairs->Length;

		for (int i = 0; i < npairs; i++)
		{
			CBMD::CMatPolyPair^ mppair = mdchunk->m_MatPolyPairs[i];
			IGLGeometry^ glg = gcnew CGLG_BMDPart(scale, mppair);

			glg->Toggle(true);

			if (glg->IsTranslucent())
				m_TSkyboxGeometry->Add(glg);
			else
				m_OSkyboxGeometry->Add(glg);
		}
	}
}

void EditorMainForm::BuildLevelGeometry()
{
	int nchunks = m_LevelModel->m_ModelChunks->Length;

	float scale = (float)m_LevelModel->m_ScaleFactor;

	for (int c = 0; c < nchunks; c++)
	{
		CBMD::CModelChunk^ mdchunk = m_LevelModel->m_ModelChunks[c];
		int npairs = mdchunk->m_MatPolyPairs->Length;

		tvModelBlocks->Nodes->Insert(c, String::Format("[{0}] {1}", mdchunk->m_ID, mdchunk->m_Name));
		tvModelBlocks->Nodes[c]->Checked = true;

		for (int i = 0; i < npairs; i++)
		{
			CBMD::CMatPolyPair^ mppair = mdchunk->m_MatPolyPairs[i];
			IGLGeometry^ glg = gcnew CGLG_BMDPart(scale, mppair);
			
			tvModelBlocks->Nodes[c]->Nodes->Insert(i, String::Format("[{0}] {1}", mppair->m_ID, mppair->m_Name));
			tvModelBlocks->Nodes[c]->Nodes[i]->Tag = glg;
			tvModelBlocks->Nodes[c]->Nodes[i]->Checked = true;

			if (glg->IsTranslucent())
				m_TLevelGeometry->Add(glg);
			else
				m_OLevelGeometry->Add(glg);
		}
	}
}

void EditorMainForm::BuildGeometry()
{
	m_OSkyboxGeometry = gcnew List<IGLGeometry^>();
	m_TSkyboxGeometry = gcnew List<IGLGeometry^>();
	m_OLevelGeometry = gcnew List<IGLGeometry^>();
	m_TLevelGeometry = gcnew List<IGLGeometry^>();

	BuildSkyboxGeometry();
	BuildLevelGeometry();
	
	m_ObjCubesList = gcnew Dictionary<UInt32,CGLG_ObjList^>();

	for (int i = 0; i < 8; i++)
	{
		CGLG_ObjList^ list0 = gcnew CGLG_ObjList(0, i, m_StdObjects[i]);
		CGLG_ObjList^ list1 = gcnew CGLG_ObjList(1, i, m_SimpleObjects[i]);
		CGLG_ObjList^ list2 = gcnew CGLG_ObjList(2, i, m_ExtraObjects[i]);

		m_ObjCubesList->Add(i<<24, list0);
		m_ObjCubesList->Add(0x10000000 | (i<<24), list1);
		m_ObjCubesList->Add(0x20000000 | (i<<24), list2);
	}

	m_SelCube = gcnew CGLG_ObjSelCube(0x00FFFFFF);
	m_HoverCube = gcnew CGLG_ObjSelCube(0x0080FFFF);

	for each (KeyValuePair<UInt32,CGLG_ObjList^>^ ol in m_ObjCubesList)
		ol->Value->SetLayerMask(m_LayerMask);

	GL::LineWidth(2.f);
	GL::PointSize(15.f);
}


void EditorMainForm::Render3DMap()
{
	GL::MatrixMode(MatrixMode::Modelview);

	GL::LoadIdentity();
	GL::MultMatrix(m_CamMatrix);

	// render pass 1 - render stuff with fake colors
	// to be used for picking detection

	GL::ClearColor(1.f, 1.f, 1.f, 1.f);
	GL::Clear(ClearBufferMask::ColorBufferBit | ClearBufferMask::DepthBufferBit | ClearBufferMask::StencilBufferBit);

	GL::Disable(EnableCap::AlphaTest);
	GL::Disable(EnableCap::Blend);
	GL::Disable(EnableCap::Dither);
	GL::Disable(EnableCap::LineSmooth);
	GL::Disable(EnableCap::PolygonSmooth);
	GL::BindTexture(TextureTarget::Texture2D, 0);

	GL::EnableClientState(ArrayCap::VertexArray);
	GL::EnableClientState(ArrayCap::ColorArray);
	GL::EnableClientState(ArrayCap::TextureCoordArray);

	for each (IGLGeometry^ glg in m_OLevelGeometry)
		glg->RenderForPicking();
	for each (KeyValuePair<UInt32,CGLG_ObjList^>^ ol in m_ObjCubesList)
		ol->Value->RenderForPicking();

	GL::Flush();
	GL::ReadPixels(m_MouseCoords.X-1, glc3DView->Height-m_MouseCoords.Y+1, 3, 3, PixelFormat::Bgra, PixelType::UnsignedByte, m_PickingFrameBuffer);

	// render pass 2 - render geometry with its actual color/texture/blahblah

	GL::ClearColor(0.f, 0.f, 0.125f, 1.f);
	GL::Clear(ClearBufferMask::ColorBufferBit | ClearBufferMask::DepthBufferBit | ClearBufferMask::StencilBufferBit);

	GL::Enable(EnableCap::AlphaTest);
	GL::Enable(EnableCap::Blend);
	GL::Enable(EnableCap::Dither);

	//GL::PolygonMode(MaterialFace::FrontAndBack, PolygonMode::Line);

	GL::LoadIdentity();
	GL::MultMatrix(m_SkyboxMatrix);

	for each (IGLGeometry^ glg in m_OSkyboxGeometry)
		glg->RenderActually();

	GL::LoadIdentity();
	GL::MultMatrix(m_CamMatrix);

	for each (IGLGeometry^ glg in m_OLevelGeometry)
		glg->RenderActually();
	for each (KeyValuePair<UInt32,CGLG_ObjList^>^ ol in m_ObjCubesList)
		ol->Value->RenderActually();
	
	GL::LoadIdentity();
	GL::MultMatrix(m_SkyboxMatrix);

	for each (IGLGeometry^ glg in m_TSkyboxGeometry)
		glg->RenderActually();

	GL::LoadIdentity();
	GL::MultMatrix(m_CamMatrix);

	for each (IGLGeometry^ glg in m_TLevelGeometry)
		glg->RenderActually();

	m_SelCube->RenderActually();
	m_HoverCube->RenderActually();

	GL::DisableClientState(ArrayCap::VertexArray);
	GL::DisableClientState(ArrayCap::ColorArray);
	GL::DisableClientState(ArrayCap::TextureCoordArray);

#if 0
	GL::Color4(1.f, 1.f, 1.f, 1.f);
	GL::BindTexture(TextureTarget::Texture2D, 0);

	int lolol = 0;
	array<Vector3>^ colors = {Vector3(1.f, 0.f, 0.f), Vector3(0.f, 1.f, 0.f), Vector3(0.f, 0.f, 1.f)};
	GL::Begin(BeginMode::Points);
	//GL::Begin(BeginMode::Triangles);
	for each (Vector3 vec in kcltest)
	{
		GL::Color3(colors[lolol % 3]);
		GL::Vertex3(vec);

		lolol++;
	}
	GL::End();
#endif
}

} // namespace sm64ds_editor
