#include "StdAfx.h"

using namespace System;

#include "NitroROM.h"
#include "NitroFile.h"

void CNitroFile::Decompress_LZ77(array<Byte>^% data)
{
	UInt32 header = (data[4] | (data[5] << 8) | (data[6] << 16) | (data[7] << 24));

	UInt32 xLen = (header >> 8);
	array<Byte>^ dest = gcnew array<Byte>(xLen);

	UInt32 xIn = 8;
	UInt32 xOut = 0;

	while(xLen > 0)
	{
		Byte d = data[xIn++];

		for(UInt32 i = 0; i < 8; i++)
		{
			if(d & 0x80)
			{
				UInt16 stuff = ((data[xIn] << 8) | data[xIn + 1]);
				xIn += 2;

				UInt32 len = (stuff >> 12) + 3;
				UInt32 offset = (stuff & 0xFFF);
				UInt32 windowOffset = (xOut - offset - 1);

				for(UInt32 j = 0; j < len; j++)
				{
					dest[xOut++] = dest[windowOffset++];

					xLen--;
					if(xLen == 0)
					{
						Array::Resize(data, dest->Length);
						dest->CopyTo(data, 0);
						delete dest;
						return;
					}
				}
			}
			else
			{
				dest[xOut++] = data[xIn++];

				xLen--;
				if(xLen == 0)
				{
					Array::Resize(data, dest->Length);
					dest->CopyTo(data, 0);
					delete dest;
					return;
				}
			}

			d <<= 1;
		}
	}

	Array::Resize(data, dest->Length);
	dest->CopyTo(data, 0);
	delete dest;
}

CNitroFile::CNitroFile(CNitroROM^ rom, UInt16 id)
{
	m_ROM = rom;
	m_ID = id;
	m_Data = m_ROM->ExtractFile(m_ID);

	// note: files that were originally LZ77 compressed, with a LZ77 tag at their beginning,
	// don't need to be recompressed before they are reinserted. Atleast not in SM64DS.
	if (Read32(0) == 0x37375A4C)
		Decompress_LZ77(m_Data);
}

CNitroFile::~CNitroFile()
{
	delete m_Data;
}

void CNitroFile::SaveChanges()
{
	// TODO: consider LZ77 recompression someday?

	m_ROM->ReinsertFile(m_ID, m_Data);
}
