#include "StdAfx.h"

using namespace System;

#include "NitroROM.h"
#include "NitroOverlay.h"

// inspired from http://nsmb-editor.googlecode.com/svn/trunk/NSMBe4/ROM.cs - DecompressOverlay()
void CNitroOverlay::Decompress(array<Byte>^% data)
{
	int size = data->Length;

	UInt32 param1, decomp_len;
	param1 = data[size-8] | (data[size-7]<<8) | (data[size-6]<<16) | (data[size-5]<<24);
	decomp_len = size + (data[size-4] | (data[size-3]<<8) | (data[size-2]<<16) | (data[size-1]<<24));

	int inpos = size - (param1 >> 24) - 1;
	int limit = size - (param1 & 0x00FFFFFF);
	int outpos = decomp_len - 1;

	Array::Resize(data, decomp_len);

	for (;;)
	{
		if (inpos <= limit) break;
		Byte blockctl = data[inpos--];
		if (inpos <= limit) break;

		bool done = false;

		for (int i = 0; i < 8; i++)
		{
			if (blockctl & 0x80)
			{
				if (inpos <= limit) { done=true; break; }

				UInt16 stuff = data[inpos-1] | (data[inpos]<<8);
				inpos -= 2;
				int wdisp = (stuff & 0x0FFF) + 2;
				int wsize = (stuff >> 12) + 2;

				for (int j = wsize; j >= 0; j--)
				{
					data[outpos] = data[outpos + wdisp + 1];
					outpos--;
				}
			}
			else
			{
				if (inpos <= limit) { done=true; break; }
				data[outpos--] = data[inpos--];
			}

			blockctl <<= 1;
		}

		if (done) break;
	}
}

CNitroOverlay::CNitroOverlay(CNitroROM^ rom, UInt32 ID)
{
	m_ROM = rom; 
	m_ID = ID;

	bool autorw = !m_ROM->CanRW();
	if (autorw) m_ROM->BeginRW();

	m_OVTEntryAddr = m_ROM->GetOverlayEntryOffset(m_ID);
	m_FileID = m_ROM->GetFileIDFromOverlayID(m_ID);

	m_RAMAddr = m_ROM->Read32(m_OVTEntryAddr + 0x04);
	Byte flags = m_ROM->Read8(m_OVTEntryAddr + 0x1F);

	m_Data = m_ROM->ExtractFile(m_FileID);
	if (flags & 0x01)
		Decompress(m_Data);

	if (autorw) m_ROM->EndRW();
}

CNitroOverlay::~CNitroOverlay()
{
	delete m_Data;
}

void CNitroOverlay::SaveChanges()
{
	bool autorw = !m_ROM->CanRW();
	if (autorw) m_ROM->BeginRW();

	// reinsert file data
	m_ROM->ReinsertFile(m_FileID, m_Data);
	Update();

	// fix overlay size
	m_ROM->Write32(m_OVTEntryAddr + 0x08, (m_Data->Length + 3) & ~3);

	// tweak the overlay table entry
	Byte flags = m_ROM->Read8(m_OVTEntryAddr + 0x1F);
	flags &= 0xFE; // [Treeki] disable compression :)
	m_ROM->Write8(m_OVTEntryAddr + 0x1F, flags);

	if (autorw) m_ROM->EndRW();
}
