#include "StdAfx.h"

using namespace System;
using namespace System::IO;
using namespace System::Collections;
using namespace System::Collections::Generic;
using namespace System::Text;
using namespace System::Text::RegularExpressions;
using namespace System::Globalization;

#include "NitroROM.h"
#include "NitroOverlay.h"

#define ROM_PATCH_VERSION 1

void CNitroROM::DecompressBinary(array<Byte>^% data)
{
	int size = data->Length;

	UInt32 param1, decomp_len;
	param1 = data[size-8] | (data[size-7]<<8) | (data[size-6]<<16) | (data[size-5]<<24);
	decomp_len = size + (data[size-4] | (data[size-3]<<8) | (data[size-2]<<16) | (data[size-1]<<24));

	int inpos = size - (param1 >> 24) - 1;
	int limit = size - (param1 & 0x00FFFFFF);
	int outpos = decomp_len - 1;

	Array::Resize(data, decomp_len);

	for (;;)
	{
		if (inpos <= limit) break;
		Byte blockctl = data[inpos--];
		if (inpos <= limit) break;

		bool done = false;

		for (int i = 0; i < 8; i++)
		{
			if (blockctl & 0x80)
			{
				if (inpos <= limit) { done=true; break; }

				UInt16 stuff = data[inpos-1] | (data[inpos]<<8);
				inpos -= 2;
				int wdisp = (stuff & 0x0FFF) + 2;
				int wsize = (stuff >> 12) + 2;

				for (int j = wsize; j >= 0; j--)
				{
					data[outpos] = data[outpos + wdisp + 1];
					outpos--;
				}
			}
			else
			{
				if (inpos <= limit) { done=true; break; }
				data[outpos--] = data[inpos--];
			}

			blockctl <<= 1;
		}

		if (done) break;
	}
}

UInt16 CNitroROM::CalcCRC16(UInt32 offset, UInt16 len)
{
	UInt32 crc = 0xFFFF;
	array<UInt16>^ val = {0xC0C1, 0xC181, 0xC301, 0xC601, 0xCC01, 0xD801, 0xF001, 0xA001};

	m_FileStream->Position = offset;
	for(UInt32 i = 0; i < len; i++)
	{
		crc = crc ^ m_BinReader->ReadByte();

		for(UInt32 j = 0; j < 8; j++)
		{
			bool carry = false;
			if (crc & 0x1) carry = true;

			crc >>= 1;

			if (carry)
				crc = crc ^ (val[j] << (7-j));
		}
	}

	return (UInt16)crc;
}

CNitroROM::CNitroROM(String^ path)
	: m_Path(path)
	, m_CanRW(false)
{
	BeginRW();

	m_FileStream->Position = 0x0C;
	UInt32 gamecode = m_BinReader->ReadUInt32();
	m_FileStream->Position = 0x1E;
	Byte romversion = m_BinReader->ReadByte();

	switch (gamecode)
	{
	case 0x454D5341: // ASME / USA
		if (romversion == 0x01)
		{
			m_LevelTableOffset = 0x8000 + 0x88ACC;
			m_ObjBanksTableOffset = 0x8000 + 0x6C384;
			m_ActorPtrTableOffset = 0x8000 + 0x8716C;
			m_FileTableOffset = 0x11244;
			m_FileTableLength = 1824;
		}
		else
		{
			m_LevelTableOffset = 0x8000 + 0x87DB0; 
			m_ObjBanksTableOffset = 0x8000 + 0x6B664;
			m_ActorPtrTableOffset = 0x8000 + 0x86450;
			m_FileTableOffset = 0x1123C;
			m_FileTableLength = 1824;
		}
		break;

	case 0x504D5341: // ASMP / EUR
		m_LevelTableOffset = 0x8000 + 0x8A208;
		m_ObjBanksTableOffset = 0x8000 + 0x6D998;
		m_ActorPtrTableOffset = 0x8000 + 0x88864;
		m_FileTableOffset = 0x13098;
		m_FileTableLength = 2058;
		break;

	default:
		throw gcnew Exception("Unknown ROM version. Tell Mega-Mario about it.");
		EndRW();
		return;
	}

	m_FileStream->Position = 0x40;
	FNTOffset = m_BinReader->ReadUInt32();
	FNTSize = m_BinReader->ReadUInt32();
	FATOffset = m_BinReader->ReadUInt32();
	FATSize = m_BinReader->ReadUInt32();
	OVTOffset = m_BinReader->ReadUInt32();
	OVTSize = m_BinReader->ReadUInt32();
	// no need to bother about ARM7 overlays... there's none in SM64DS

	m_FileStream->Position = 0x80;
	m_UsedSize = m_BinReader->ReadUInt32();

	m_FileStream->Position = FNTOffset+6;
	UInt16 numdirs = m_BinReader->ReadUInt16();
	UInt32 numfiles = FATSize/8;

	m_DirEntries = gcnew array<CDirEntry^>(numdirs);
	m_FileEntries = gcnew array<CFileEntry^>(numfiles);

	m_FileStream->Position = FATOffset;
	for (UInt32 f = 0; f < numfiles; f++)
	{
		UInt32 start = m_BinReader->ReadUInt32();
		UInt32 end = m_BinReader->ReadUInt32();

		m_FileEntries[f] = gcnew CFileEntry((UInt16)f, start, end-start);
	}

	m_DirEntries[0] = gcnew CDirEntry(0xF000, 0, "");

	UInt32 tableoffset = FNTOffset;
	for (int d = 0; d < numdirs; d++)
	{
		m_FileStream->Position = tableoffset;
		UInt32 subtableoffset = FNTOffset + m_BinReader->ReadUInt32();
		UInt16 first_fileid = m_BinReader->ReadUInt16();
		UInt16 cur_fileid = first_fileid;

		m_FileStream->Position = subtableoffset;
		for (;;)
		{
			Byte type_len = m_BinReader->ReadByte();

			if (type_len == 0x00) break;
			else if (type_len > 0x80)
			{
				array<wchar_t>^ _name = m_BinReader->ReadChars(type_len & 0x7F);
				UInt16 dirid = m_BinReader->ReadUInt16();

				m_DirEntries[dirid-0xF000] = gcnew CDirEntry(dirid, 0xF000+d, gcnew String(_name));
			}
			else if (type_len < 0x80)
			{
				array<wchar_t>^ _name = m_BinReader->ReadChars(type_len & 0x7F);

				m_FileEntries[cur_fileid]->ParentID = 0xF000+d;
				m_FileEntries[cur_fileid]->Name = gcnew String(_name);
				cur_fileid++;
			}
		}

		tableoffset += 8;
	}

	for each (CDirEntry^ de in m_DirEntries)
	{
		if (de->ParentID > 0xF000)
			de->FullName = String::Concat(m_DirEntries[de->ParentID-0xF000]->FullName, "/", de->Name);
		else
			de->FullName = de->Name;
	}

	for each (CFileEntry^ fe in m_FileEntries)
	{
		if (fe->ParentID > 0xF000)
			fe->FullName = String::Concat(m_DirEntries[fe->ParentID-0xF000]->FullName, "/", fe->Name);
		else
			fe->FullName = fe->Name;
	}

	UInt32 numoverlays = OVTSize/0x20;
	m_OverlayEntries = gcnew array<COverlayEntry^>(numoverlays);

	for (UInt32 i = 0; i < numoverlays; i++)
	{
		m_FileStream->Position = OVTOffset + (i*0x20);

		UInt32 ovlid = m_BinReader->ReadUInt32();
		m_FileStream->Position += 0x14;
		UInt16 fileid = m_BinReader->ReadUInt16();

		m_OverlayEntries[ovlid] = gcnew COverlayEntry(OVTOffset + (i*0x20), ovlid, fileid);
	}
	
	EndRW();


	m_ObjectNames = gcnew array<String^>(512);
	for (int i = 0; i < 512; i++)
		m_ObjectNames[i] = "(unknown)";

	try
	{
		FileStream^ fs = File::OpenRead("./obj_list.txt");
		if (fs->CanRead)
		{
			StreamReader^ sr = gcnew StreamReader(fs);

			String^ curline;
			Regex^ lineregex = gcnew Regex("0x([\\dabcdef]+) == (.*?) \\(0x([\\dabcdef]+)\\)");

			while ((curline = sr->ReadLine()) != nullptr)
			{
				Match^ stuff = lineregex->Match(curline);
				m_ObjectNames->SetValue(stuff->Groups[2]->Value, int::Parse(stuff->Groups[1]->Value, NumberStyles::HexNumber));
			}
			 
			delete lineregex;
			delete sr;
		}
		delete fs;
	}
	catch (Exception^) {}
}

CNitroROM::~CNitroROM()
{
	for each (COverlayEntry^ oe in m_OverlayEntries)
		delete oe;

	for each (CFileEntry^ fe in m_FileEntries)
		delete fe;

	for each (CDirEntry^ de in m_DirEntries)
		delete de;

	delete m_OverlayEntries;
	delete m_FileEntries;
	delete m_DirEntries;
}

bool CNitroROM::NeedsPatch()
{
	m_FileStream->Position = 0x1FC;
	return (m_BinReader->ReadUInt32() != ROM_PATCH_VERSION);
}

void CNitroROM::Patch()
{
	// write patch version to the end of the ROM header
	// this place should normally be zero and unused :)
	m_FileStream->Position = 0x1FC;
	m_BinWriter->Write(UInt32(ROM_PATCH_VERSION));

	// read the ARM9 binary, decompress it and write it back
	m_FileStream->Position = 0x2C;
	UInt32 oldsize = m_BinReader->ReadUInt32() - 0x4000;

	m_FileStream->Position = 0x8000;
	array<Byte>^ binary = m_BinReader->ReadBytes(oldsize);
	DecompressBinary(binary);

	int newsize = (binary->Length + 3) & ~3;
	int delta = newsize - oldsize;

	m_FileStream->Position = 0x8000 + oldsize;
	array<Byte>^ tomove = m_BinReader->ReadBytes(m_UsedSize - (0x8000 + oldsize));
	m_FileStream->Position = 0x8000 + newsize;
	m_BinWriter->Write(tomove);
	delete tomove;

	m_FileStream->Position = 0x8000;
	m_BinWriter->Write(binary);

	// patch the loader code so that it doesn't try to decompress
	// the already decompressed binary
	UInt32 nop = 0xE1A00000;
	m_FileStream->Position = 0x484C;
	m_BinWriter->Write(nop);
	m_BinWriter->Write(nop);
	m_BinWriter->Write(nop);

	int fixstart = 0x8000 + oldsize;

	// fix our internal variables to keep 'em reliable
	if (FNTOffset >= fixstart) FNTOffset += delta;
	if (FATOffset >= fixstart) FATOffset += delta;
	if (m_UsedSize >= fixstart) m_UsedSize += delta;

	for each (CFileEntry^ _fe in m_FileEntries)
	{
		if (_fe->Offset >= fixstart)
			_fe->Offset += delta;
	}

	if (OVTOffset >= fixstart)
	{
		OVTOffset += delta;

		for each (COverlayEntry^ oe in m_OverlayEntries)
			oe->EntryOffset += delta;
	}

	// fix the actual ROM
	UInt32 temp;
#define FIX(_offset) \
	m_FileStream->Position = (_offset); \
	temp = m_BinReader->ReadUInt32(); \
	if (temp >= fixstart) \
	{ \
		temp += delta; \
		m_FileStream->Position = (_offset); \
		m_BinWriter->Write(temp); \
	}

	FIX(0x20);
	FIX(0x30);
	FIX(0x40);
	FIX(0x48);
	FIX(0x50);
	FIX(0x58);
	FIX(0x68);
	FIX(0x70);
	FIX(0x74);
	FIX(0x80);
	FIX(0x160);

	for (int i = 0; i < (FATSize / 4); i++)
	{
		FIX(FATOffset + (i*4));
	}

#undef FIX

	// fix the ARM9 binary size - important
	UInt32 newarm9size = 0x4000 + newsize;
	m_FileStream->Position = 0x2C;
	m_BinWriter->Write(newarm9size);

	// fix the header CRC16... we never know :P
	// as an example NO$GBA won't load the ROM if this CRC16 is wrong
	UInt16 hcrc = CalcCRC16(0, 0x15E);
	m_FileStream->Position = 0x15E;
	m_BinWriter->Write(hcrc);

	delete binary;
}

void CNitroROM::LoadFileTable()
{
	m_FileTable = gcnew array<UInt16>(m_FileTableLength);

	CNitroOverlay^ ovl0 = gcnew CNitroOverlay(this, 0);
	for (UInt32 i = 0; i < m_FileTableLength; i++)
	{
		UInt32 str_offset = ovl0->Read32(m_FileTableOffset + (i*4)) - ovl0->GetRAMAddr();
		String^ fname = ovl0->ReadString(str_offset, 0);
		m_FileTable[i] = GetFileIDFromName(fname);
	}
	delete ovl0;
}

UInt16 CNitroROM::GetFileIDFromName(String^ name)
{
	for each (CFileEntry^ fe in m_FileEntries)
	{
		if (fe->FullName == name)
			return fe->ID;
	}

	throw gcnew Exception(String::Format("File '{0}' not found in ROM {1}", name, m_Path));
	return 0;
}

UInt16 CNitroROM::GetFileIDFromOverlayID(UInt32 ovlid)
{
	return m_OverlayEntries[ovlid]->FileID;
}

array<Byte>^ CNitroROM::ExtractFile(UInt16 fileid)
{
	bool autorw = !m_CanRW;
	if (autorw) BeginRW();

	CFileEntry^ fe = m_FileEntries[fileid];

	m_FileStream->Position = fe->Offset;
	array<Byte>^ data = m_BinReader->ReadBytes(fe->Size);

	if (autorw) EndRW();
	return data;
}

void CNitroROM::ReinsertFile(UInt16 fileid, array<Byte>^ data)
{
	bool autorw = !m_CanRW;
	if (autorw) BeginRW();

	int datalength = (data->Length + 3) & ~3;

	CFileEntry^ fe = m_FileEntries[fileid];

	UInt32 fileend = fe->Offset + fe->Size;
	int delta = datalength - fe->Size;

	// move data that comes after the file
	m_FileStream->Position = fileend;
	array<Byte>^ tomove = m_BinReader->ReadBytes(m_UsedSize - fileend);
	m_FileStream->Position = fileend + delta;
	m_BinWriter->Write(tomove);
	delete tomove;

	// write the new data for the file
	m_FileStream->Position = fe->Offset;
	m_BinWriter->Write(data);
	fe->Size = datalength;

	// fix our internal variables to keep 'em reliable
	if (FNTOffset >= fileend) FNTOffset += delta;
	if (FATOffset >= fileend) FATOffset += delta;
	if (m_UsedSize >= fileend) m_UsedSize += delta;

	for each (CFileEntry^ _fe in m_FileEntries)
	{
		if (_fe->Offset >= fileend)
			_fe->Offset += delta;
	}

	if (OVTOffset >= fileend)
	{
		OVTOffset += delta;

		for each (COverlayEntry^ oe in m_OverlayEntries)
			oe->EntryOffset += delta;
	}

	// fix the actual ROM
	UInt32 temp;
#define FIX(_offset) \
	m_FileStream->Position = (_offset); \
	temp = m_BinReader->ReadUInt32(); \
	if (temp >= fileend) \
	{ \
		temp += delta; \
		m_FileStream->Position = (_offset); \
		m_BinWriter->Write(temp); \
	}

	FIX(0x20);
	FIX(0x30);
	FIX(0x40);
	FIX(0x48);
	FIX(0x50);
	FIX(0x58);
	FIX(0x68);
	FIX(0x70);
	FIX(0x74);
	FIX(0x80);
	FIX(0x160);

	for (int i = 0; i < (FATSize / 4); i++)
	{
		FIX(FATOffset + (i*4));
	}

	// fix the header CRC16... we never know :P
	// as an example NO$GBA won't load the ROM if this CRC16 is wrong
	UInt16 hcrc = CalcCRC16(0, 0x15E);
	m_FileStream->Position = 0x15E;
	m_BinWriter->Write(hcrc);

#undef FIX

	if (autorw) EndRW();
}

array<UInt32>^ CNitroROM::FixTables(int level, UInt32 fixstart, int delta)
{
	bool autorw = !m_CanRW;
	if (autorw) BeginRW();

	// hm so Nintendo wants to play that game? let's show 'em we can hardcode stuff too :P
	// TODO: [247] WATER_SUIKOMI?
	//       [248] HANACHAN?
	array<String^>^ specActors = {
		"", // 0 
		"338|339|342|343",
		"36|37|38",
		"",
		"34|35",
		"292|293|294",
		"40|41|219",
		"42|44|45|50|51|52|53|54",
		"56|57|58|60|242|313",
		"62",
		"63|257|295|355", // 10
		"259|356", 
		"213|325|326|327|328",
		"67|68|221",
		"70|71|73|74|75|76|77|80|82|83|243",
		"84",
		"85|86",
		"87|88|162|163",
		"89|90|91|246",
		"92|93|217|258|275",
		"", // 20
		"94|95|96|97|98|99|100|101",
		"102|103|267|268",
		"302|303|304|305",
		"106|107|228",
		"104|105",
		"",
		"119|120|121|122",
		"123|125|126|127|129|130|132",
		"",
		"", // 30
		"314",
		"",
		"",
		"",
		"134|135|136|137",
		"301",
		"139|140|141|142|145|148",
		"",
		"151|152|153|156",
		"", // 40
		"",
		"",
		"",
		"174|175",
		"",
		"",
		"195",
		"173",
		"",
		"182", // 50
		""
	};

	// fix the level header address
	UInt32 ptr = Read32(m_LevelTableOffset + (level*4));
	if (ptr >= fixstart)
	{
		ptr += delta;
		Write32(m_LevelTableOffset + (level*4), ptr);
	}

	// fix addresses of actors specific to the level (if any)
	array<String^>^ tofix = specActors[level]->Split(gcnew array<Char> {'|'});
	array<UInt32>^ ret = gcnew array<UInt32>(tofix->Length);
	int i = 0;

	for each (String^ _aid in tofix)
	{
		UInt32 aid = UInt32::Parse(_aid);
		UInt32 ptr = Read32(m_ActorPtrTableOffset + (aid*4));
		ret[i++] = ptr;

		if (ptr >= fixstart)
		{
			ptr += delta;
			Write32(m_ActorPtrTableOffset + (aid*4), ptr);
		}
	}

	if (autorw) EndRW();
	return ret;
}

void CNitroROM::BeginRW()
{
	if (m_CanRW) return;

	m_FileStream = File::Open(m_Path, FileMode::Open, FileAccess::ReadWrite);
	m_BinReader = gcnew BinaryReader(m_FileStream, Text::Encoding::ASCII);
	m_BinWriter = gcnew BinaryWriter(m_FileStream, Text::Encoding::ASCII);
	m_CanRW = true;
}

void CNitroROM::EndRW()
{
	if (!m_CanRW) return;

	m_CanRW = false;
	delete m_BinWriter;
	delete m_BinReader;
	delete m_FileStream;
}
