#pragma once
// not so generic NitroROM class
// crafted especially for SM64DS but contains generic stuff
// that can be used for other ROMs

public ref class CNitroROM
{
public:
	CNitroROM(String^ path);
	~CNitroROM();

	String^ GetPath() { return m_Path; }

	bool NeedsPatch();
	void Patch();

	void LoadFileTable();

	UInt16 GetFileIDFromName(String^ name);
	UInt16 GetFileIDFromOverlayID(UInt32 ovlid);
	UInt16 GetFileIDFromInternalID(UInt16 intid) { return m_FileTable[intid]; }

	UInt32 GetFileOffset(UInt16 fileid) { return m_FileEntries[fileid]->Offset; }
	UInt32 GetOverlayEntryOffset(UInt32 ovlid) { return m_OverlayEntries[ovlid]->EntryOffset; }

	array<Byte>^ ExtractFile(UInt16 fileid);
	void ReinsertFile(UInt16 fileid, array<Byte>^ data);

	void BeginRW();
	void EndRW();
	bool CanRW() { return m_CanRW; }

	Byte Read8(UInt32 addr) { m_FileStream->Position = addr; return m_BinReader->ReadByte(); }
	UInt16 Read16(UInt32 addr) { m_FileStream->Position = addr; return m_BinReader->ReadUInt16(); }
	UInt32 Read32(UInt32 addr) { m_FileStream->Position = addr; return m_BinReader->ReadUInt32(); }
	array<Byte>^ ReadBlock(UInt32 addr, UInt32 size) { m_FileStream->Position = addr; return m_BinReader->ReadBytes(size); }

	void Write8(UInt32 addr, Byte value) { m_FileStream->Position = addr; m_BinWriter->Write(value); }
	void Write16(UInt32 addr, UInt16 value) { m_FileStream->Position = addr; m_BinWriter->Write(value); }
	void Write32(UInt32 addr, UInt32 value) { m_FileStream->Position = addr; m_BinWriter->Write(value); }
	void WriteBlock(UInt32 addr, array<Byte>^ data) { m_FileStream->Position = addr; m_BinWriter->Write(data); }

	array<UInt32>^ FixTables(int level, UInt32 fixstart, int delta);

	UInt32 GetLevelAddress(int level) { return Read32(m_LevelTableOffset + (level*4)); }
	array<Byte>^ GetLevelObjBanks(int level) { return ReadBlock(m_ObjBanksTableOffset + (level*7), 7); }

	void SetLevelAddress(int level, UInt32 addr) { Write32(m_LevelTableOffset + (level*4), addr); }
	void SetLevelObjBanks(int level, array<Byte>^ banks) { WriteBlock(m_ObjBanksTableOffset + (level*7), banks); }

	String^ GetObjectName(int id) { return m_ObjectNames[id]; }

private:
	String^ m_Path;

	bool m_CanRW;
	IO::Stream^ m_FileStream;
	IO::BinaryReader^ m_BinReader;
	IO::BinaryWriter^ m_BinWriter;

	UInt32 m_LevelTableOffset;
	UInt32 m_ObjBanksTableOffset;
	UInt32 m_ActorPtrTableOffset;
	UInt32 m_FileTableOffset, m_FileTableLength;
	array<UInt16>^ m_FileTable;

	UInt32 m_UsedSize;

	UInt32 FNTOffset, FNTSize;
	UInt32 FATOffset, FATSize;
	UInt32 OVTOffset, OVTSize;

	ref class CDirEntry
	{
	public:
		CDirEntry() {}
		CDirEntry(UInt16 _id, UInt16 _parentid, String^ _name)
			: ID(_id), ParentID(_parentid), Name(_name) {}
		~CDirEntry() {}

		UInt16 ID;
		UInt16 ParentID;
		String^ Name;
		String^ FullName;
	};

	ref class CFileEntry
	{
	public:
		CFileEntry() {}
		CFileEntry(UInt16 _id, UInt32 _offset, UInt32 _size)
			: ID(_id), ParentID(0), Name(""), Offset(_offset), Size(_size) {}
		~CFileEntry() {}

		UInt16 ID;
		UInt16 ParentID;
		String^ Name;
		String^ FullName;
		UInt32 Offset;
		UInt32 Size;
	};

	ref class COverlayEntry
	{
	public:
		COverlayEntry(UInt32 _entryoffset, UInt32 _id, UInt16 _fileid)
			: EntryOffset(_entryoffset), ID(_id), FileID(_fileid) {}
		~COverlayEntry() {}

		UInt32 EntryOffset;
		UInt32 ID;
		UInt16 FileID;
	};

	array<CDirEntry^>^ m_DirEntries;
	array<CFileEntry^>^ m_FileEntries;
	array<COverlayEntry^>^ m_OverlayEntries;

	void DecompressBinary(array<Byte>^% data);
	UInt16 CalcCRC16(UInt32 offset, UInt16 len);

	array<String^>^ m_ObjectNames;
};

public ref class INitroROMBlock
{
public:
	virtual Byte Read8(UInt32 addr) { return m_Data[addr]; }
	virtual UInt16 Read16(UInt32 addr) { return (UInt16)(m_Data[addr] | (m_Data[addr+1]<<8)); }
	virtual UInt32 Read32(UInt32 addr) { return (UInt32)(m_Data[addr] | (m_Data[addr+1]<<8) | (m_Data[addr+2]<<16) | (m_Data[addr+3]<<24)); }

	virtual array<Byte>^ ReadBlock(UInt32 addr, UInt32 len)
	{
		array<Byte>^ ret = gcnew array<Byte>(len);
		Array::Copy(m_Data, (int)addr, ret, 0, (int)len);
		return ret;
	}

	// reads a string until the specified length or until a null byte
	// if length is zero, no length limit is applied
	virtual String^ ReadString(UInt32 addr, UInt32 len)
	{
		Collections::Generic::List<Char>^ templist = gcnew Collections::Generic::List<Char>(256);
		for (UInt32 i = 0; ; i++)
		{
			if ((len > 0) && (i >= len)) break;

			Char ch = (Char)m_Data[addr+i];
			if (ch == 0) break;

			templist->Add(ch);
		}

		String^ result = gcnew String(templist->ToArray());
		delete templist;
		return result;
	}

	virtual void Write8(UInt32 addr, Byte value) { m_Data[addr] = value; }
	virtual void Write16(UInt32 addr, UInt16 value) { m_Data[addr] = value & 0xFF; m_Data[addr+1] = value >> 8; }
	virtual void Write32(UInt32 addr, UInt32 value) { m_Data[addr] = value & 0xFF; m_Data[addr+1] = (value>>8) & 0xFF; m_Data[addr+2] = (value>>16) & 0xFF; m_Data[addr+3] = value>>24; }

	virtual void WriteBlock(UInt32 addr, array<Byte>^ data)
	{
		if (((int)addr + data->Length) > m_Data->Length)
			Array::Resize(m_Data, addr + data->Length);

		Array::Copy(data, 0, m_Data, addr, data->Length);
	}

//private:
	CNitroROM^ m_ROM;
	array<Byte>^ m_Data;
};
