#pragma once

using namespace System::Windows::Forms;
using namespace System::Globalization;

public ref class CLevelSettings
{
public:
	CLevelSettings(int id, CNitroOverlay^ ovl, UInt32 offset)
	{
		m_LevelID = id;
		m_Overlay = ovl;
		m_Offset = offset;

		Background = ovl->Read8(offset+0x18) >> 4;
		ObjBanks = ovl->m_ROM->GetLevelObjBanks(id);
	}
	~CLevelSettings() {}

	void SaveChanges()
	{
		m_Overlay->Write8(m_Offset+0x18, (Background<<4) | 0x0F);
		m_Overlay->m_ROM->SetLevelObjBanks(m_LevelID, ObjBanks);
	}

private:
	int m_LevelID;
	CNitroOverlay^ m_Overlay;
	UInt32 m_Offset;

public:
	Byte Background;
	array<Byte>^ ObjBanks;
};



public ref class IAnyObject
{
public:
	virtual bool IsPositioned() { return false; }
	virtual bool CanBeRotated() { return false; }
	virtual int GetObjType() { return -666; }
	virtual String^ GetDesc() { return "LOL"; }
	virtual int GetEntrySize() { return 0; }

	virtual void FillEditPanel(TableLayoutPanel^ panel, UInt32 arg) {}
	virtual void UpdateFromEditPanel(TableLayoutPanel^ panel, int id) {}

	virtual void SaveChanges() {}

	CNitroOverlay^ m_Overlay;
	UInt32 m_Offset;
	UInt32 m_UniqueID;
	Byte m_Layer;
};

public ref class IPositionedObject : public IAnyObject
{
public:
	virtual bool IsPositioned() override { return true; }

	float XPos, YPos, ZPos;
	float YRot;
};

public ref class CStdObject : public IPositionedObject
{
public:
	CStdObject(CNitroOverlay^ ovl, UInt32 offset, Byte layer, UInt32 uid, bool isnew)
	{
		m_Overlay = ovl;
		m_Offset = offset;
		m_Layer = layer;
		m_UniqueID = ((layer & 0xF) << 24) | (uid & 0x00FFFFFF);

		if (isnew)
		{
			ObjectID = 0x001F;
			XPos = 0.f;
			YPos = 0.f;
			ZPos = 0.f;
			YRot = 0.f;
			Param1 = 0xFF00;
			Param2 = 0;
			Param3 = 0;
		}
		else
		{
			ObjectID = ovl->Read16(offset);
			XPos = (float)((Int16)ovl->Read16(offset+2)) / 1000.f;
			YPos = (float)((Int16)ovl->Read16(offset+4)) / 1000.f;
			ZPos = (float)((Int16)ovl->Read16(offset+6)) / 1000.f;
			YRot = (float)((Int16)ovl->Read16(offset+10)) / 4096.f;
			Param1 = ovl->Read16(offset+14);
			Param2 = ovl->Read16(offset+8);
			Param3 = ovl->Read16(offset+12);
		}
	}
	~CStdObject() {}

	virtual bool CanBeRotated() override { return true; }
	virtual int GetObjType() override { return 0; }
	virtual String^ GetDesc() override { return String::Format("{0:X4} - {1}", ObjectID, m_Overlay->m_ROM->GetObjectName(ObjectID)); }
	virtual int GetEntrySize() override { return 16; }

	virtual void FillEditPanel(TableLayoutPanel^ panel, UInt32 arg) override
	{
		if (arg == 0)
		{
			((ComboBox^)panel->Controls[1])->SelectedIndex = ObjectID;
			panel->Controls[11]->Text = String::Format("{0:X4}", Param1);
			panel->Controls[13]->Text = String::Format("{0:X4}", Param2);
			panel->Controls[15]->Text = String::Format("{0:X4}", Param3);
		}
		
		((NumericUpDown^)panel->Controls[3])->Value = Decimal(XPos);
		((NumericUpDown^)panel->Controls[5])->Value = Decimal(YPos);
		((NumericUpDown^)panel->Controls[7])->Value = Decimal(ZPos);
		((NumericUpDown^)panel->Controls[9])->Value = Decimal(YRot*22.5f);
	}

	virtual void UpdateFromEditPanel(TableLayoutPanel^ panel, int id) override
	{
		switch (id)
		{
		case 1:
			{
				int oid = ((ComboBox^)panel->Controls[1])->SelectedIndex;
				if (oid == -1)
				{
					if (!int::TryParse(((ComboBox^)panel->Controls[1])->Text, NumberStyles::HexNumber, nullptr, oid))
						oid = -1;
				}

				if (oid != -1) ObjectID = (UInt16)oid;
			}
			break;
		case 3: XPos = (float)((NumericUpDown^)panel->Controls[3])->Value; break;
		case 5: YPos = (float)((NumericUpDown^)panel->Controls[5])->Value; break;
		case 7: ZPos = (float)((NumericUpDown^)panel->Controls[7])->Value; break;
		case 9: YRot = (float)((NumericUpDown^)panel->Controls[9])->Value / 22.5f; break;
		case 11:
			{
				UInt16 param;
				if (UInt16::TryParse(panel->Controls[11]->Text, NumberStyles::HexNumber, nullptr, param)) 
					Param1 = param;
			}
			break;
		case 13:
			{
				UInt16 param;
				if (UInt16::TryParse(panel->Controls[13]->Text, NumberStyles::HexNumber, nullptr, param)) 
					Param2 = param;
			}
			break;
		case 15:
			{
				UInt16 param;
				if (UInt16::TryParse(panel->Controls[15]->Text, NumberStyles::HexNumber, nullptr, param)) 
					Param3 = param;
			}
			break;
		}
	}

	virtual void SaveChanges() override
	{
		m_Overlay->Write16(m_Offset, ObjectID);
		m_Overlay->Write16(m_Offset+2, (UInt16)((Int16)(XPos * 1000.f)));
		m_Overlay->Write16(m_Offset+4, (UInt16)((Int16)(YPos * 1000.f)));
		m_Overlay->Write16(m_Offset+6, (UInt16)((Int16)(ZPos * 1000.f)));
		m_Overlay->Write16(m_Offset+10, (UInt16)((Int16)(YRot * 4096.f)));
		m_Overlay->Write16(m_Offset+14, Param1);
		m_Overlay->Write16(m_Offset+8, Param2);
		m_Overlay->Write16(m_Offset+12, Param3);
	}

	UInt16 ObjectID;
	UInt16 Param1, Param2, Param3;
};

public ref class CSimpleObject : public IPositionedObject
{
public:
	CSimpleObject(CNitroOverlay^ ovl, UInt32 offset, Byte layer, UInt32 uid, bool isnew)
	{
		m_Overlay = ovl;
		m_Offset = offset;
		m_Layer = layer;
		m_UniqueID = 0x10000000 | ((layer & 0xF) << 24) | (uid & 0x00FFFFFF);

		if (isnew)
		{
			ObjectID = 0x29;
			Param1 = 0xF;
			Param2 = 0x7;
			XPos = 0.f;
			YPos = 0.f;
			ZPos = 0.f;
			YRot = 0.f;
		}
		else
		{
			UInt16 shizz = ovl->Read16(offset);
			ObjectID = shizz & 0x1FF;
			Param1 = (shizz >> 9) & 0xF;
			Param2 = (shizz >> 13) & 0x7;
			XPos = (float)((Int16)ovl->Read16(offset+2)) / 1000.f;
			YPos = (float)((Int16)ovl->Read16(offset+4)) / 1000.f;
			ZPos = (float)((Int16)ovl->Read16(offset+6)) / 1000.f;
			YRot = 0.f;
		}
	}
	~CSimpleObject() {}

	virtual int GetObjType() override { return 1; }
	virtual String^ GetDesc() override { return String::Format("{0:X4} - {1}", ObjectID, m_Overlay->m_ROM->GetObjectName(ObjectID)); }
	virtual int GetEntrySize() override { return 8; }

	virtual void FillEditPanel(TableLayoutPanel^ panel, UInt32 arg) override
	{
		if (arg == 0)
		{
			((ComboBox^)panel->Controls[1])->SelectedIndex = ObjectID;
			panel->Controls[9]->Text = String::Format("{0:X4}", Param1);
			panel->Controls[11]->Text = String::Format("{0:X4}", Param2);
		}
		
		((NumericUpDown^)panel->Controls[3])->Value = Decimal(XPos);
		((NumericUpDown^)panel->Controls[5])->Value = Decimal(YPos);
		((NumericUpDown^)panel->Controls[7])->Value = Decimal(ZPos);
	}

	virtual void UpdateFromEditPanel(TableLayoutPanel^ panel, int id) override
	{
		switch (id)
		{
		case 1:
			{
				int oid = ((ComboBox^)panel->Controls[1])->SelectedIndex;
				if (oid == -1)
				{
					if (!int::TryParse(((ComboBox^)panel->Controls[1])->Text, NumberStyles::HexNumber, nullptr, oid))
						oid = -1;
				}

				if (oid != -1) ObjectID = (UInt16)oid;
			}
			break;
		case 3: XPos = (float)((NumericUpDown^)panel->Controls[3])->Value; break;
		case 5: YPos = (float)((NumericUpDown^)panel->Controls[5])->Value; break;
		case 7: ZPos = (float)((NumericUpDown^)panel->Controls[7])->Value; break;
		case 9:
			{
				UInt16 param;
				if (UInt16::TryParse(panel->Controls[9]->Text, NumberStyles::HexNumber, nullptr, param)) 
					Param1 = param;
			}
			break;
		case 11:
			{
				UInt16 param;
				if (UInt16::TryParse(panel->Controls[11]->Text, NumberStyles::HexNumber, nullptr, param)) 
					Param2 = param;
			}
			break;
		}
	}

	virtual void SaveChanges() override
	{
		UInt16 shizz = (ObjectID & 0x1FF) | ((Param1 & 0xF) << 9) | ((Param2 & 0x7) << 13);
		m_Overlay->Write16(m_Offset, shizz);
		m_Overlay->Write16(m_Offset+2, (UInt16)((Int16)(XPos * 1000.f)));
		m_Overlay->Write16(m_Offset+4, (UInt16)((Int16)(YPos * 1000.f)));
		m_Overlay->Write16(m_Offset+6, (UInt16)((Int16)(ZPos * 1000.f)));
	}

	UInt16 ObjectID;
	UInt16 Param1, Param2;
};

public ref class CExtraObject : public IPositionedObject
{
public:
	CExtraObject(CNitroOverlay^ ovl, UInt32 offset, Byte layer, UInt32 uid)
	{
		m_Overlay = ovl;
		m_Offset = offset;
		m_Layer = layer;
		m_UniqueID = 0x20000000 | ((layer & 0xF) << 24) | (uid & 0x00FFFFFF);

		ObjectID = ovl->Read16(offset);
		XPos = (float)((Int16)ovl->Read16(offset+2)) / 1000.f;
		YPos = (float)((Int16)ovl->Read16(offset+4)) / 1000.f;
		ZPos = (float)((Int16)ovl->Read16(offset+6)) / 1000.f;
		YRot = (float)((Int16)ovl->Read16(offset+10)) / 4096.f;
		Param1 = ovl->Read16(offset+14);
		Param2 = ovl->Read16(offset+8);
		Param3 = ovl->Read16(offset+12);
	}
	~CExtraObject() {}

	virtual bool CanBeRotated() override { return true; }
	virtual int GetObjType() override { return 2; }
	virtual String^ GetDesc() override { return String::Format("{0:X4} - {1}", ObjectID, m_Overlay->m_ROM->GetObjectName(ObjectID)); }
	virtual int GetEntrySize() override { return 16; }

	virtual void FillEditPanel(TableLayoutPanel^ panel, UInt32 arg) override
	{
		if (arg == 0)
		{
			((ComboBox^)panel->Controls[1])->SelectedIndex = ObjectID;
			panel->Controls[11]->Text = String::Format("{0:X4}", Param1);
			panel->Controls[13]->Text = String::Format("{0:X4}", Param2);
			panel->Controls[15]->Text = String::Format("{0:X4}", Param3);
		}
		
		((NumericUpDown^)panel->Controls[3])->Value = Decimal(XPos);
		((NumericUpDown^)panel->Controls[5])->Value = Decimal(YPos);
		((NumericUpDown^)panel->Controls[7])->Value = Decimal(ZPos);
		((NumericUpDown^)panel->Controls[9])->Value = Decimal(YRot*22.5f);
	}

	virtual void UpdateFromEditPanel(TableLayoutPanel^ panel, int id) override
	{
		switch (id)
		{
		case 1:
			{
				int oid = ((ComboBox^)panel->Controls[1])->SelectedIndex;
				if (oid == -1)
				{
					if (!int::TryParse(((ComboBox^)panel->Controls[1])->Text, NumberStyles::HexNumber, nullptr, oid))
						oid = -1;
				}

				if (oid != -1) ObjectID = (UInt16)oid;
			}
			break;
		case 3: XPos = (float)((NumericUpDown^)panel->Controls[3])->Value; break;
		case 5: YPos = (float)((NumericUpDown^)panel->Controls[5])->Value; break;
		case 7: ZPos = (float)((NumericUpDown^)panel->Controls[7])->Value; break;
		case 9: YRot = (float)((NumericUpDown^)panel->Controls[9])->Value / 22.5f; break;
		case 11:
			{
				UInt16 param;
				if (UInt16::TryParse(panel->Controls[11]->Text, NumberStyles::HexNumber, nullptr, param)) 
					Param1 = param;
			}
			break;
		case 13:
			{
				UInt16 param;
				if (UInt16::TryParse(panel->Controls[13]->Text, NumberStyles::HexNumber, nullptr, param)) 
					Param2 = param;
			}
			break;
		case 15:
			{
				UInt16 param;
				if (UInt16::TryParse(panel->Controls[15]->Text, NumberStyles::HexNumber, nullptr, param)) 
					Param3 = param;
			}
			break;
		}
	}

	virtual void SaveChanges() override
	{
		m_Overlay->Write16(m_Offset, ObjectID);
		m_Overlay->Write16(m_Offset+2, (UInt16)((Int16)(XPos * 1000.f)));
		m_Overlay->Write16(m_Offset+4, (UInt16)((Int16)(YPos * 1000.f)));
		m_Overlay->Write16(m_Offset+6, (UInt16)((Int16)(ZPos * 1000.f)));
		m_Overlay->Write16(m_Offset+10, (UInt16)((Int16)(YRot * 4096.f)));
		m_Overlay->Write16(m_Offset+14, Param1);
		m_Overlay->Write16(m_Offset+8, Param2);
		m_Overlay->Write16(m_Offset+12, Param3);
	}

	UInt16 ObjectID;
	UInt16 Param1, Param2, Param3;
};
